'''Thermo RAW file reading implementation using the pure .NET
RawFileReader library released in 2017.

This module provides :class:`ThermoRawLoader`, a :class:`~.RandomAccessScanSource`
implementation.

Depends upon the ``pythonnet`` project which provides the :mod:`clr`
module, enabling nearly seamless interoperation with the Common Language
Runtime.

The public interface of this module should be identical to
:mod:`ms_deisotope.data_source.thermo_raw`.
'''

import sys
import os

from collections import OrderedDict

import numpy as np

from pyteomics.auxiliary import unitfloat

from ms_peak_picker import PeakSet, PeakIndex, simple_peak
from ms_deisotope.data_source.common import (
    PrecursorInformation, ChargeNotProvided, Scan,
    ActivationInformation, MultipleActivationInformation,
    IsolationWindow, ScanDataSource, ScanEventInformation,
    ScanAcquisitionInformation, ScanWindow, RandomAccessScanSource)
from ms_deisotope.data_source.metadata.activation import (
    supplemental_term_map, dissociation_methods_map)
from ms_deisotope.data_source._thermo_helper import (
    _InstrumentMethod, ThermoRawScanPtr, FilterString,
    _make_id, _id_template, _RawFileMetadataLoader, analyzer_map)


_DEFAULT_DLL_PATH = os.path.join(
    os.path.dirname(
        os.path.realpath(__file__)),
    "_vendor",
    "ThermoRawFileReader_3_0_41",
    "Libraries")


# late binding imports

Business = None
_RawFileReader = None
clr = None
NullReferenceException = Exception
Marshal = None
IntPtr = None
Int64 = None


def is_thermo_raw_file(path):
    '''Detect whether or not the file referenced by ``path``
    is a Thermo RAW file.

    Parameters
    ----------
    path: :class:`str`
        The path to test

    Returns
    -------
    :class:`bool`:
        Whether or not the file is a Thermo RAW file.
    '''
    if not _test_dll_loaded():
        try:
            register_dll()
        except ImportError:
            return False
    try:
        source = _RawFileReader.RawFileReaderAdapter.FileFactory(path)
        source.SelectInstrument(Business.Device.MS, 1)
        return True
    except NullReferenceException:   # pylint: disable=broad-except
        return False


def infer_reader(path):
    '''If the file referenced by ``path`` is a Thermo RAW
    file, return the callable (:class:`ThermoRawLoader`) to
    open it, otherwise raise an exception.

    Parameters
    ----------
    path: :class:`str`
        The path to test

    Returns
    -------
    :class:`type`:
        The type to use to open the file

    Raises
    ------
    :class:`ValueError`:
        If the file is not a Thermo RAW file
    '''
    if is_thermo_raw_file(path):
        return ThermoRawLoader
    raise ValueError("Not Thermo Raw File")


def determine_if_available():
    '''Checks whether or not the COM-based Thermo
    RAW file reading feature is available.

    This is done by attempting to instantiate the
    COM-provided object, which queries the Windows
    registry for the MSFileReader.dll.

    Returns
    -------
    :class:`bool`:
        Whether or not the feature is enabled.
    '''
    try:
        return _register_dll([_DEFAULT_DLL_PATH])
    except (OSError, ImportError):
        return False


def _register_dll(search_paths=None):
    '''Start the Common Language Runtime interop service by importing
    the :mod:`clr` module from Pythonnet, and then populate the global
    names referring to .NET entities, and finally attempt to locate the
    ThermoRawFileReader DLLs by searching alogn ``search_paths``.

    Parameters
    ----------
    search_paths: list
        The paths to check along for the ThermoRawFileReader DLL bundle.

    Returns
    -------
    :class:`bool`:
        Whether or not the .NET library successfully loaded
    '''
    if search_paths is None:
        search_paths = []
    global _RawFileReader, Business, clr, NullReferenceException   # pylint: disable=global-statement
    global Marshal, IntPtr, Int64   # pylint: disable=global-statement
    if _test_dll_loaded():
        return True
    try:
        import clr  # pylint: disable=redefined-outer-name
        from System import NullReferenceException  # pylint: disable=redefined-outer-name
        clr.AddReference("System.Runtime")
        clr.AddReference("System.Runtime.InteropServices")
        from System import IntPtr, Int64  # pylint: disable=redefined-outer-name
        from System.Runtime.InteropServices import Marshal  # pylint: disable=redefined-outer-name
    except ImportError:
        return False
    for path in search_paths:
        sys.path.append(path)
        try:
            clr.AddReference('ThermoFisher.CommonCore.RawFileReader')
            clr.AddReference('ThermoFisher.CommonCore.Data')
        except OSError:
            continue
        try:
            import ThermoFisher.CommonCore.Data.Business as Business  # pylint: disable=redefined-outer-name
            import ThermoFisher.CommonCore.RawFileReader as _RawFileReader  # pylint: disable=redefined-outer-name
        except ImportError:
            continue
    return _test_dll_loaded()


def register_dll(search_paths=None):
    '''Register the location of the Thermo RawFileReader DLL bundle with
    the Common Language Runtime interop system and load the .NET symbols
    used by this feature.

    Parameters
    ----------
    search_paths: list
        The paths to check along for the ThermoRawFileReader DLL bundle.

    '''
    if search_paths is None:
        search_paths = []
    search_paths = list(search_paths)
    search_paths.append(_DEFAULT_DLL_PATH)
    loaded = _register_dll(search_paths)
    if not loaded:
        msg = '''The ThermoFisher.CommonCore libraries could not be located and loaded.'''
        raise ImportError(msg)


def _test_dll_loaded():
    return _RawFileReader is not None


def _copy_double_array(src):
    '''A quick and dirty implementation of the fourth technique shown in
    https://mail.python.org/pipermail/pythondotnet/2014-May/001525.html for
    copying a .NET Array[Double] to a NumPy ndarray[np.float64] via a raw
    memory copy.

    ``int_ptr_tp`` must be an integer type that can hold a pointer. On Python 2
    this is :class:`long`, and on Python 3 it is :class:`int`.
    '''
    dest = np.empty(len(src), dtype=np.float64)
    Marshal.Copy(
        src, 0,
        IntPtr.__overloads__[Int64](dest.__array_interface__['data'][0]),
        len(src))
    return dest


class RawReaderInterface(ScanDataSource):
    ''':class:`~.ScanDataSource` implementation for Thermo's RawFileReader API.
    Not intended for direct instantiation.
    '''

    def _scan_arrays(self, scan):
        scan_number = scan.scan_number + 1
        stats = self._source.GetScanStatsForScanNumber(scan_number)
        segscan = self._source.GetSegmentedScanFromScanNumber(scan_number, stats)
        mzs = _copy_double_array(segscan.Positions)
        inten = _copy_double_array(segscan.Intensities)
        return mzs, inten

    def _pick_peaks_vendor(self, scan, *args, **kwargs):
        scan_info = Business.Scan.FromFile(self._source, scan.scan_number + 1)
        if scan_info.HasCentroidStream:
            stream = self._source.GetCentroidStream(scan.scan_number + 1, 0)
            mzs = stream.Masses
            intens = stream.Intensities
            peaks = PeakSet([simple_peak(mzs[i], intens[i], 0.001) for i in range(len(mzs))])
            peaks.reindex()
            arrays = self._scan_arrays(scan)
            return PeakIndex(arrays[0], arrays[1], peaks)
        else:
            raise NotImplementedError()

    def _scan_id(self, scan):
        scan_number = scan.scan_number
        return _make_id(scan_number + 1)

    def _is_profile(self, scan):
        return not self._source.IsCentroidScanFromScanNumber(
            scan.scan_number + 1)

    def _polarity(self, scan):
        filter_string = self._filter_string(scan)
        return filter_string.data['polarity']

    def _scan_title(self, scan):
        return "%s %r" % (self._scan_id(scan), self._filter_string(scan))

    def _filter_string(self, scan):
        scan_number = scan.scan_number
        return FilterString(self._source.GetFilterForScanNumber(scan_number + 1).Filter)

    def _scan_index(self, scan):
        scan_number = scan.scan_number
        return scan_number

    def _scan_time(self, scan):
        scan_number = scan.scan_number
        return self._source.RetentionTimeFromScanNumber(scan_number + 1)

    def _ms_level(self, scan):
        scan_number = scan.scan_number
        f = self._source.GetFilterForScanNumber(scan_number + 1)
        return f.MSOrder

    def _isolation_window(self, scan):
        scan_number = scan.scan_number
        filt = self._source.GetFilterForScanNumber(scan_number + 1)
        seq_index = filt.MSOrder - 2
        width = filt.GetIsolationWidth(seq_index)
        offset = filt.GetIsolationWidthOffset(seq_index)
        precursor_mz = filt.GetMass(seq_index)
        return IsolationWindow(width, precursor_mz + offset, width)

    def _trailer_values(self, scan):
        scan_number = scan.scan_number
        trailers = self._source.GetTrailerExtraInformation(scan_number + 1)
        return OrderedDict(zip([label.strip(":") for label in trailers.Labels], trailers.Values))

    def _infer_precursor_scan_number(self, scan):
        precursor_scan_number = None
        last_index = self._scan_index(scan) - 1
        current_level = self._ms_level(scan)
        i = 0
        while last_index >= 0 and i < 100:
            prev_scan = self.get_scan_by_index(last_index)
            if prev_scan.ms_level >= current_level:
                last_index -= 1
            else:
                precursor_scan_number = prev_scan._data.scan_number
                break
            i += 1
        return precursor_scan_number

    def _precursor_information(self, scan):
        scan_number = scan.scan_number
        filt = self._source.GetFilterForScanNumber(scan_number + 1)
        precursor_mz = filt.GetMass(filt.MSOrder - 2)
        trailers = self._trailer_values(scan)
        _precursor_mz = float(trailers.get("Monoisotopic M/Z", 0))
        if _precursor_mz > 0:
            precursor_mz = _precursor_mz

        # imitate proteowizard's firmware bug correction
        isolation_window = self._isolation_window(scan)
        if (isolation_window.upper + isolation_window.lower) / 2 <= 2.0:
            if (isolation_window.target - 3.0 > precursor_mz) or (isolation_window.target + 2.5 < precursor_mz):
                precursor_mz = isolation_window.target
        elif precursor_mz not in isolation_window:
            precursor_mz = isolation_window.target
        charge = int(trailers.get("Charge State", 0))
        if charge == 0:
            charge = ChargeNotProvided
        inten = 0
        precursor_scan_number = None
        if precursor_scan_number is None:
            last_index = self._scan_index(scan) - 1
            current_level = self._ms_level(scan)
            i = 0
            while last_index >= 0 and i < 100:
                prev_scan = self.get_scan_by_index(last_index)
                if prev_scan.ms_level >= current_level:
                    last_index -= 1
                else:
                    precursor_scan_number = prev_scan._data.scan_number
                    break
                i += 1
        if precursor_scan_number is not None:
            precursor_scan_id = self.get_scan_by_index(precursor_scan_number).id
        return PrecursorInformation(
            precursor_mz, inten, charge, precursor_scan_id,
            source=self, product_scan_id=self._scan_id(scan))

    def _get_scan_segment(self, scan):
        trailer = self._trailer_values(scan)
        try:
            return int(trailer['Scan Segment'])
        except KeyError:
            return 1

    def _get_scan_event(self, scan):
        trailer = self._trailer_values(scan)
        try:
            return int(trailer['Scan Event'])
        except KeyError:
            return 1

    def _activation(self, scan):
        filter_string = self._filter_string(scan)
        tandem_sequence = filter_string.get("tandem_sequence")
        # If the tandem sequence exists, the last entry is the most recent tandem acquisition.
        # It will list contain one or more activation types. Alternatively, multiple activations
        # of the same precursor may exist in the list as separate events in the tandem sequence.
        if tandem_sequence is not None:
            activation_event = tandem_sequence[-1]
            activation_type = list(activation_event.get("activation_type"))
            has_supplemental_activation = filter_string.get("supplemental_activation")

            if activation_type is not None:
                energy = list(activation_event.get("activation_energy"))
                if len(tandem_sequence) > 1:
                    prev_event = tandem_sequence[-2]
                    # Merge previous tandem sequences of the same precursor
                    if abs(prev_event['isolation_mz'] - activation_event['isolation_mz']) < 1e-3:
                        activation_type = list(prev_event.get("activation_type")) + activation_type
                        energy = list(prev_event.get("activation_energy")) + energy
                        has_supplemental_activation = True

                if has_supplemental_activation and len(activation_type) > 1:
                    activation_type.append(supplemental_term_map[
                        dissociation_methods_map[activation_type[-1]]])
                if len(activation_type) == 1:
                    return ActivationInformation(activation_type[0], energy[0])
                else:
                    return MultipleActivationInformation(activation_type, energy)
        return None

    def _acquisition_information(self, scan):
        fline = self._filter_string(scan)
        event = self._get_scan_event(scan)
        trailer_extras = self._trailer_values(scan)
        traits = {
            'preset scan configuration': event,
            'filter string': fline,
        }
        event = ScanEventInformation(
            self._scan_time(scan),
            injection_time=unitfloat(trailer_extras.get('Ion Injection Time (ms)', 0.0), 'millisecond'),
            window_list=[ScanWindow(
                fline.get("scan_window")[0], fline.get("scan_window")[1])], traits=traits)
        return ScanAcquisitionInformation("no combination", [event])

    def _instrument_configuration(self, scan):
        fline = self._filter_string(scan)
        try:
            confid = self._analyzer_to_configuration_index[analyzer_map[fline.data.get("analyzer")]]
            return self._instrument_config[confid]
        except KeyError:
            return None

    def _annotations(self, scan):
        fline = self._filter_string(scan)
        trailer_extras = self._trailer_values(scan)
        annots = {
            "filter string": fline,
        }
        microscans = trailer_extras.get("Micro Scan Count")
        if microscans is not None:
            annots['[Thermo Trailer Extra]Micro Scan Count'] = float(microscans)
        scan_segment = trailer_extras.get("Scan Segment")
        if scan_segment is not None:
            annots['[Thermo Trailer Extra]Scan Segment'] = int(scan_segment)
        scan_event = trailer_extras.get("Scan Event")
        if scan_event is not None:
            annots['[Thermo Trailer Extra]Scan Event'] = int(scan_event)
        mono_mz = float(trailer_extras.get("Monoisotopic M/Z", 0))
        if mono_mz is not None and mono_mz > 0:
            annots['[Thermo Trailer Extra]Monoisotopic M/Z'] = mono_mz
        hcd_ev = trailer_extras.get('HCD Energy eV')
        if hcd_ev is not None and hcd_ev > 0:
            annots['[Thermo Trailer Extra]HCD Energy eV'] = hcd_ev
        hcd_energies = trailer_extras.get('HCD Energy')
        if hcd_energies is not None and hcd_energies:
            annots['[Thermo Trailer Extra]HCD Energy'] = hcd_energies
        return annots


class ThermoRawLoader(RawReaderInterface, RandomAccessScanSource, _RawFileMetadataLoader):
    '''Reads scans from Thermo Fisher RAW files directly. Provides both iterative and
    random access.
    '''

    def __init__(self, source_file, _load_metadata=True, **kwargs):
        if not _test_dll_loaded():
            register_dll()
        self._source = _RawFileReader.RawFileReaderAdapter.FileFactory(source_file)
        self._source.SelectInstrument(Business.Device.MS, 1)
        self.source_file = source_file
        self._producer = None
        self._scan_type_index = dict()
        self.make_iterator()
        self.initialize_scan_cache()
        self._first_scan_time = self.get_scan_by_index(0).scan_time
        self._last_scan_time = self.get_scan_by_index(
            self._source.RunHeaderEx.LastSpectrum - 1).scan_time
        self._index = self._pack_index()

        if _load_metadata:
            self._method = self._parse_method()
            self._build_scan_type_index()
            self._get_instrument_info()

    def _has_ms1_scans(self):
        if self._scan_type_index:
            return 1 in self._scan_type_index
        else:
            # metadata has not been loaded so best to assume there is
            return True

    def _has_msn_scans(self):
        if self._scan_type_index:
            return max(self._scan_type_index) > 1
        else:
            # metadata has not been loaded so best to assume there is
            return True

    def has_msn_scans(self):
        return self._has_msn_scans()

    def has_ms1_scans(self):
        return self._has_ms1_scans()

    def __reduce__(self):
        return self.__class__, (self.source_file, False)

    @property
    def index(self):
        '''Accesses the scan index

        Returns
        -------
        :class:`collections.OrderedDict`
            Maps scan ID to index
        '''
        return self._index

    def __len__(self):
        return len(self.index)

    def __repr__(self):
        return "ThermoRawLoader(%r)" % (self.source_file)

    def close(self):
        '''Close the underlying file reader.
        '''
        if self._source is not None:
            self._source.Close()
            self._source = None

    def __del__(self):
        self.close()

    def reset(self):
        self.make_iterator(None)
        self.initialize_scan_cache()

    def _pack_index(self):
        index = OrderedDict()
        for sn in range(self._source.RunHeaderEx.FirstSpectrum - 1,
                        self._source.RunHeaderEx.LastSpectrum):
            index[_make_id(sn)] = sn
        return index

    def _get_instrument_model_name(self):
        return self._source.GetInstrumentData().Model;

    def _get_instrument_serial_number(self):
        return self._source.GetInstrumentData().SerialNumber

    def _parse_method(self):
        try:
            method_count = self._source.InstrumentMethodsCount
            if method_count == 0:
                return _InstrumentMethod('')
            # the data acquisition method should be the last method
            return _InstrumentMethod(self._source.GetInstrumentMethod(method_count - 1))
        except NullReferenceException:   # pylint: disable=broad-except
            return _InstrumentMethod('')

    def _scan_time_to_scan_number(self, scan_time):
        scan_number = self._source.ScanNumberFromRetentionTime(scan_time) - 1
        return scan_number

    def get_scan_by_index(self, index):
        """Retrieve the scan object for the specified scan index.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        index: int
            The index to get the scan for

        Returns
        -------
        Scan
        """
        scan_number = int(index)
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = ThermoRawScanPtr(scan_number)
            if not package.validate(self):
                raise IndexError(index)
            scan = Scan(package, self)
            self._scan_cache[scan_number] = scan
            return scan

    def get_scan_by_id(self, scan_id):
        """Retrieve the scan object for the specified scan id.

        If the scan object is still bound and in memory somewhere,
        a reference to that same object will be returned. Otherwise,
        a new object will be created.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        Scan
        """
        scan_number = int(str(scan_id).replace(_id_template, '')) - 1
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = ThermoRawScanPtr(scan_number)
            if not package.validate(self):
                raise KeyError(str(scan_id))
            scan = Scan(package, self)
            self._scan_cache[scan_number] = scan
            return scan

    def get_scan_by_time(self, time):
        """Retrieve the scan object for the specified scan time.

        This internally calls :meth:`get_scan_by_id` which will
        use its cache.

        Parameters
        ----------
        time : float
            The time to get the nearest scan from

        Returns
        -------
        Scan
        """
        if time < self._first_scan_time:
            time = self._first_scan_time
        elif time > self._last_scan_time:
            time = self._last_scan_time
        scan_number = self._scan_time_to_scan_number(time)
        try:
            return self._scan_cache[scan_number]
        except KeyError:
            package = ThermoRawScanPtr(scan_number)
            scan = Scan(package, self)
            self._scan_cache[scan_number] = scan
            return scan

    def start_from_scan(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        '''Reconstruct an iterator which will start from the scan matching one of ``scan_id``,
        ``rt``, or ``index``. Only one may be provided.

        After invoking this method, the iterator this object wraps will be changed to begin
        yielding scan bunchs (or single scans if ``grouped`` is ``False``).

        Arguments
        ---------
        scan_id: str, optional
            Start from the scan with the specified id.
        rt: float, optional
            Start from the scan nearest to specified time (in minutes) in the run. If no
            exact match is found, the nearest scan time will be found, rounded up.
        index: int, optional
            Start from the scan with the specified index.
        require_ms1: bool, optional
            Whether the iterator must start from an MS1 scan. True by default.
        grouped: bool, optional
            whether the iterator should yield scan bunches or single scans. True by default.
        '''
        if scan_id is not None:
            scan_number = int(str(scan_id).replace(_id_template, '')) - 1
        elif index is not None:
            scan_number = int(index)
        elif rt is not None:
            scan_number = self._scan_time_to_scan_number(rt)
        if require_ms1:
            start_index = scan_number
            while start_index != 0:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            scan_number = start_index
        iterator = self._make_pointer_iterator(start_index=scan_number)
        if grouped:
            self._producer = self._scan_group_iterator(iterator)
        else:
            self._producer = self._single_scan_iterator(iterator)
        return self

    def _make_scan_index_producer(self, start_index=None, start_time=None):
        if start_index is not None:
            return range(start_index, self._source.RunHeaderEx.LastSpectrum - 1)
        elif start_time is not None:
            start_index = self._scan_time_to_scan_number(start_time)
            while start_index != 0:
                scan = self.get_scan_by_index(start_index)
                if scan.ms_level > 1:
                    start_index -= 1
                else:
                    break
            return range(start_index, self._source.RunHeaderEx.LastSpectrum - 1)
        else:
            return range(0, self._source.RunHeaderEx.LastSpectrum - 1)

    def _make_pointer_iterator(self, start_index=None, start_time=None):
        iterator = self._make_scan_index_producer(start_index, start_time)
        for i in iterator:
            yield ThermoRawScanPtr(i)

    def _make_default_iterator(self):
        return self._make_pointer_iterator()

    def _make_cache_key(self, scan):
        return scan._data.scan_number

    def next(self):
        return next(self._producer)



import logging

try:
    from ms_deisotope.data_source._vendor.AgilentD import (
        AgilentDLoader, register_dll_dir, AgilentDScanPtr,
        AgilentDDataInterface,
        log as _api_logger)
    from comtypes import COMError

    comtypes_logger = logging.getLogger("comtypes")
    comtypes_logger.setLevel("INFO")
    _api_logger.setLevel("INFO")

    def is_agilent_d_dir(path):
        try:
            AgilentDLoader(path)
            return True
        except (WindowsError, IOError, ImportError, COMError):
            return False

    def infer_reader(path):
        if is_agilent_d_dir(path):
            return AgilentDLoader
        raise ValueError("Not Agilent .d Directory")

    def determine_if_available():
        try:
            AgilentDLoader.create_com_object()
            return True
        except (WindowsError, COMError):
            return False
except ImportError as e:
    message = str(e)

    def is_agilent_d_dir(path):
        return False

    def infer_reader(path):
        raise ValueError(message)

    def register_dll_dir(paths):
        print("no-op: %s" % (message,))
        return False

    def determine_if_available():
        print("no-op: %s" % (message,))
        return False

    class AgilentDLoader(object):
        def __init__(self, *args, **kwargs):
            raise ImportError(message)



import operator
import logging

from collections import defaultdict

from .subgraph import ConnectedSubgraph
from .intervals import SpanningMixin, IntervalTreeNode
from ..utils import Base, TargetedDeconvolutionResultBase
from ..task import LogUtilsMixin

logger = logging.getLogger("deconvolution_scan_processor")


def ident(x):
    return x


class PeakNode(Base):
    """
    Represent a single FittedPeak and the junction of multiple
    IsotopicFitRecords which depend uponit

    Attributes
    ----------
    links : dict
        Mapping IsotopicFitRecords which depend upon this peak to those fit's scores
    peak : ms_peak_picker.FittedPeak
        The peak being depended upon
    """

    def __init__(self, peak, links=None):
        if links is None:
            links = {}
        self.peak = peak
        self.links = links
        self._hash = hash(self.peak)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        try:
            return self.peak == other.peak
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other

    def __contains__(self, fit):
        return fit in self.links

    def __repr__(self):
        return "PeakNode(%s, %s)" % (self.peak, self.links)


class DependenceCluster(SpanningMixin):
    """
    Represent a cluster of isotopic fits which are overlapping

    Attributes
    ----------
    dependencies : set
        The set of all fits which are codependent
    start: float
        The smallest mz of all elements of this cluster
    end: float
        The largest mz of all elements of this cluster
    maximize: bool
        Whether the objective is to maximize or minimize the
        isotopic fit score
    """

    def __init__(self, parent=None, dependencies=None, maximize=True):
        if parent is None:
            parent = self
        if dependencies is None:
            dependencies = []
        else:
            dependencies = sorted(dependencies, key=lambda x: x.score, reverse=maximize)
        self.parent = parent
        self.dependencies = dependencies
        self.rank = 0
        self.maximize = maximize
        self._reset()

    def _reset(self):
        self.start = self._start()
        self.end = self._end()
        self.best_fit = self._best_fit()

    def add(self, fit):
        """
        Adds a new IsotopicFitRecord to this cluster, and ensures the sorted
        property still holds

        Parameters
        ----------
        fit : IsotopicFitRecord
            The fit being added to the record

        """
        self.dependencies.append(fit)
        self.dependencies.sort(key=lambda x: x.score, reverse=self.maximize)
        self._reset()

    def disjoint_subset(self):
        graph = self.build_graph()
        return graph.find_heaviest_path()

    def build_graph(self):
        graph = ConnectedSubgraph(self.dependencies, maximize=self.maximize)
        return graph

    def _best_fit(self):
        """
        Retrieve the absolute best single fit for the cluster

        Returns
        -------
        name : IsotopicFitRecord
            The best scoring IsotopicFitRecord in this cluster
        """
        return self.dependencies[0]

    def disjoint_best_fits(self):
        """
        Compute the best set of disjoint isotopic fits spanning this cluster

        Returns
        -------
        list of IsotopicFitRecord
        """
        fit_sets = tuple(self.disjoint_subset())
        best_fits = fit_sets
        return [node.fit for node in best_fits]

    def _start(self):
        """
        Determine the first mz coordinate for members of this cluster

        Returns
        -------
        float
        """
        return min([member.experimental[0].mz for member in self.dependencies])

    def _end(self):
        """
        Determines the last mz coordinate for members of this cluster

        Returns
        -------
        float
        """
        return max([member.experimental[-1].mz for member in self.dependencies])

    def __len__(self):
        return len(self.dependencies)

    def __iter__(self):
        return iter(self.dependencies)

    def __contains__(self, fit):
        return fit in self.dependencies

    def __eq__(self, other):
        return self.dependencies == other.dependencies

    def __repr__(self):
        return "DependenceCluster(dependencies=[%s], start=%0.4f, end=%0.4f)" % (
            '\n'.join(map(str, self.dependencies[:10])), self.start, self.end)

    def __getitem__(self, i):
        return self.dependencies[i]


try:
    _has_c = True
    from ms_deisotope._c.peak_dependency_network.peak_network import (
        PeakNode, DependenceCluster, PeakDependenceGraphBase)
except ImportError:
    _has_c = False

    class PeakDependenceGraphBase(object):

        def reset(self):
            # Keep a record of all clusters from previous iterations
            self._all_clusters.extend(
                self.clusters if self.clusters is not None else [])
            self.nodes = dict()
            self.dependencies = set()
            self._interval_tree = None
            self._populate_initial_graph()

        def _populate_initial_graph(self):
            for peak in self.peaklist:
                self.nodes[peak.index] = PeakNode(peak)

        def add_fit_dependence(self, fit_record):
            '''Add the relatoinship between the experimental peaks
            in `fit_record` to the graph, expressed as a hyper-edge
            denoted by `fit_record`.

            This adds `fit_record` to :attr:`PeakNode.links` for each
            node corresponding to the :class:`~.FittedPeak` instance in
            :attr:`IsotopicFitRecord.experimental` of `fit_record`. It also
            adds `fit_record to :attr:`dependencies`

            Parameters
            ----------
            fit_record: :class:`~.IsotopicFitRecord`
            '''
            for peak in fit_record.experimental:
                # Check for null peak
                if peak.index == 0:
                    continue
                self.nodes[peak.index].links[fit_record] = fit_record.score
            self.dependencies.add(fit_record)

        def nodes_for(self, fit_record, cache=None):
            if cache is None:
                return [self.nodes[p.index] for p in fit_record.experimental if p.peak_count >= 0]
            else:
                try:
                    return cache[fit_record]
                except KeyError:
                    cache[fit_record] = value = [
                        self.nodes[p.index] for p in fit_record.experimental if p.peak_count >= 0]
                    return value


class PeakDependenceGraph(PeakDependenceGraphBase, LogUtilsMixin):
    def __init__(self, peaklist, nodes=None, dependencies=None, max_missed_peaks=1,
                 use_monoisotopic_superceded_filtering=True, maximize=True):
        if nodes is None:
            nodes = {}
        if dependencies is None:
            dependencies = set()
        self.peaklist = peaklist
        self.nodes = nodes
        self.dependencies = dependencies
        self.max_missed_peaks = max_missed_peaks
        self.use_monoisotopic_superceded_filtering = use_monoisotopic_superceded_filtering
        self.maximize = maximize
        if len(self.nodes) == 0:
            self._populate_initial_graph()
        self.clusters = None
        self._interval_tree = None
        self._solution_map = {}
        self._all_clusters = []

    def add_solution(self, key, solution):
        self._solution_map[key] = solution

    @property
    def interval_tree(self):
        '''An :class:`~.IntervalTreeNode` built over all dependency clusters that
        the :class:`PeakDependenceGraph` has seen over the course of all iterations.

        Raises
        ------
        NoIsotopicClustersError:
            When no isotopic clusters, and in turn no :class:`DependenceCluster`s are found

        Returns
        -------
        :class:`~.IntervalTreeNode`
        '''
        if self._interval_tree is None:
            # Build tree from both the current set of clusters
            # and all clusters from previous iterations
            self._interval_tree = IntervalTreeNode.build(
                self.clusters + self._all_clusters)
            if self._interval_tree is None:
                raise NoIsotopicClustersError(
                    "Could not build intervals for peak retrieval with %d clusters" % len(
                        self.clusters), self)
        return self._interval_tree

    def _find_fuzzy_solution_for(self, peak, shift=0.5):
        tree = self.interval_tree
        clusters = tree.contains_point(peak.mz + shift)
        if len(clusters) == 0:
            # No dependency clusters span this point. No signal?
            return None
        else:
            best_fits = [cluster.disjoint_best_fits() for cluster in clusters]

            acc = []
            for fits in best_fits:
                acc.extend(fits)
            best_fits = acc

            index = 0
            error = float('inf')

            for i, fit in enumerate(best_fits):
                err = abs(fit.monoisotopic_peak.mz - peak.mz)
                if err < error:
                    error = err
                    index = i
            fit = best_fits[index]
            try:
                return self._solution_map[fit]
            except KeyError:
                best_fits.sort(key=lambda f: abs(f.monoisotopic_peak.mz - peak.mz))
                # The first index is the minimum error, which the loop above found,
                # but by virtue of the KeyError getting us here, we know it is not
                # present in the solution map.
                for f in  best_fits[1:]:
                    try:
                        return self._solution_map[f]
                    except KeyError:
                        continue
                # No solution could be found
                return None

    def find_solution_for(self, peak):
        '''Find the best isotopic pattern fit which includes ``peak``
        or is close enough to ``peak`` that it might have been selected
        instead.

        Parameters
        ----------
        peak: :class:`~.FittedPeak`
            The peak to find the best fit for.

        Returns
        -------
        :class:`~.DeconvolutedPeak`
        '''
        peak_node = self.nodes[peak.index]
        tree = self.interval_tree
        clusters = tree.contains_point(peak.mz)
        if not clusters:
            return self._find_fuzzy_solution_for(peak)
        best_fits = [cluster.disjoint_best_fits() for cluster in clusters]

        acc = []
        for fits in best_fits:
            acc.extend(fits)
        best_fits = acc

        self.debug("For query m/z %f has candidates %r" % (peak.mz, best_fits))

        # Extract only fits that use the query peak
        common = tuple(set(best_fits) & set(peak_node.links))

        used_common = False
        if len(common) > 1 or not common:
            if len(common) > 1:
                self.debug("There is not exactly one solution for %r (%d)" % (peak, len(common)))
            # If there were no fits for this peak, then it may be that this peak
            # was not included in a fit. Try to find the nearest solution.
            fit = self._find_best_fit_by_weighted_distance(peak, best_fits)
        else:
            used_common = True
            fit = common[0]
        try:
            return self._solution_map[fit]
        except KeyError:
            self.debug(
                "The closest common fit was not chosen. The closest fit was common? %r" % (
                    used_common))
            candidates = list(set(best_fits) & set(self._solution_map))
            fit = self._find_best_fit_by_weighted_distance(peak, candidates)
            return self._solution_map[fit]

    def _find_best_fit_by_weighted_distance(self, peak, candidates):
        '''Locate the fit that is nearest to ``peak`` in m/z space, weighted by
        the score of the fit such that given two fits with the same m/z, the better
        scoring fit is chosen.

        Parameters
        ----------
        peak: :class:`~.FittedPeak`
            The peak to find solutions with reference to.
        candidates: :class:`list`
            A set of candidate :class:`~.IsotopicFitRecord` to compare between

        Returns
        -------
        :class:`~.IsotopicFitRecord`
        '''
        i = 0
        err = float('inf')
        for j, case in enumerate(candidates):
            case_err = abs(case.monoisotopic_peak.mz - peak.mz)
            if self.maximize:
                case_err /= case.score
            else:
                case_err /= (1. / (case.score + 1e-6))
            if case_err < err:
                i = j
                err = case_err
        fit = candidates[i]
        return fit

    def drop_fit_dependence(self, fit_record):
        '''Remove this fit from the graph, deleting all
        hyper-edges.
        '''
        for node in self.nodes_for(fit_record):
            try:
                del node.links[fit_record]
            except KeyError:
                pass

    def claimed_nodes(self):
        '''Return all nodes with at least one
        fit connecting it.

        Returns
        -------
        :class:`set`
        '''
        peaks = set()
        for fit in self.dependencies:
            for peak in fit.experimental:
                if peak.peak_count < 0:
                    continue
                peaks.add(self.nodes[peak.index])
        return peaks

    def drop_superceded_fits(self):
        '''Drop all fits where there is a better fit
        using the first fit's monoisotopic peak at the
        same charge state.
        '''
        suppressed = []
        keep = []

        if self.maximize:
            comparator = operator.lt
        else:
            comparator = operator.gt

        for dep in self.dependencies:
            mono_peak = dep.monoisotopic_peak
            if mono_peak.peak_count < 0:
                continue
            mono_peak_node = self.nodes[mono_peak.index]
            suppress = False
            for candidate in mono_peak_node.links:
                if dep.charge == candidate.charge:
                    if mono_peak in candidate.experimental:
                        if comparator(dep.score, candidate.score):
                            suppress = True
                            break
            if suppress:
                suppressed.append(dep)
            else:
                keep.append(dep)
        for dropped in suppressed:
            self.drop_fit_dependence(dropped)
        self.dependencies = set(keep)

    def best_exact_fits(self):
        '''For each distinct group of experimental peaks, retain only
        the best scoring fit using exactly those peaks.
        '''
        by_peaks = defaultdict(list)
        best_fits = []
        for fit in self.dependencies:
            by_peaks[tuple(fit.experimental)].append(fit)
        for peak_tuple, fits in by_peaks.items():
            fits = sorted(fits, key=lambda x: x.score, reverse=not self.maximize)
            for fit in fits[:-1]:
                self.drop_fit_dependence(fit)
            best_fits.append(fits[-1])
        self.dependencies = set(best_fits)

    def drop_gapped_fits(self, n=None):
        '''Discard any fit with more missed peaks than
        :attr:`max_missed_peaks`.
        '''
        if n is None:
            n = self.max_missed_peaks
        keep = []
        for dep in self.dependencies:
            if dep.missed_peaks > n:
                self.drop_fit_dependence(dep)
            else:
                keep.append(dep)
        self.dependencies = set(keep)

    def find_non_overlapping_intervals(self):
        clusters = defaultdict(set)
        self.drop_gapped_fits()
        self.best_exact_fits()
        if self.use_monoisotopic_superceded_filtering:
            self.drop_superceded_fits()

        nodes_for_cache = {}

        for seed_node in self.nodes.values():
            # This peak is depended upon by each fit in `dependencies`
            dependencies = set(seed_node.links.keys())

            if len(dependencies) == 0:
                continue

            # These fits also depend upon these other peaks, and those peaks are depended upon
            # for other fits in turn, which depend upon the assignment of this peak.
            for dep in list(dependencies):
                for node in self.nodes_for(dep, nodes_for_cache):
                    dependencies |= clusters[node]

            # Create a fresh copy to share out again to avoid eliminate possible errors of shared storage.
            # This is likely unecessary under the current implementation.
            dependencies = set(dependencies)

            # Update all co-depended nodes with the full set of all fits which depend upon them
            for dep in dependencies:
                for node in self.nodes_for(dep, nodes_for_cache):
                    clusters[node] = dependencies

        # Use an `id` keyed dictionary over these copied sets to ensure we have exactly one reference
        # to each set of inter-dependent fits, and then convert each set into an instance of `DependenceCluster`
        clusters = [DependenceCluster(dependencies=c, maximize=self.maximize) for c in {
            id(v): v for v in clusters.values() if v}.values()]
        clusters = sorted(clusters, key=operator.attrgetter("start"))
        self.clusters = clusters
        return clusters

    def __iter__(self):
        return iter(self.clusters)

    def __getitem__(self, i):
        return self.clusters[i]

    def __repr__(self):
        return "PeakDependenceNetwork(%s, %d)" % (self.peaklist, len(self.dependencies))


PeakDependenceGraph.log_with_logger(logger)


class NetworkedTargetedDeconvolutionResult(TargetedDeconvolutionResultBase):
    '''Stores the necessary information to retrieve the local optimal solution for
    a single peak for a deconvolution algorithm from the optimal fit containing
    :attr:`query_peak` in the set of disjoint best fits for the enclosing connected
    component

    Attributes
    ----------
    query_peak : FittedPeak
        The peak queried with
    solution_peak : DeconvolutedPeak
        The optimal solution peak
    '''
    def __init__(self, deconvoluter, peak, *args, **kwargs):
        super(NetworkedTargetedDeconvolutionResult, self).__init__(deconvoluter, *args, **kwargs)
        self.query_peak = peak
        self.solution_peak = None

    def _get_solution(self):
        try:
            self.solution_peak = self.deconvoluter.peak_dependency_network.find_solution_for(
                self.query_peak)
        except IndexError:
            pass

    def get(self):
        """Fetch the optimal solution after the computation has finished.

        Returns
        -------
        DeconvolutedPeak
        """
        self._get_solution()
        return self.solution_peak


class NoIsotopicClustersError(ValueError):
    def __init__(self, msg, graph=None, scan_id=None):
        ValueError.__init__(self, msg)
        self.graph = graph
        self.scan_id = scan_id



'''A collection of tools for drawing and annotating mass spectra
'''
# pragma: no cover
import math
import itertools

import numpy as np
try:
    from matplotlib import pyplot as plt, gridspec
    from ms_peak_picker.plot import draw_peaklist, draw_raw
    has_plot = True
except ImportError:
    import warnings
    warnings.warn("Could not import matplotlib, plotting tools will not work")
    pyplot = None
    gridspec = None
    has_plot = False


def _default_color_cycle():
    c = plt.rcParams['axes.prop_cycle']
    colors = c.by_key().get("color")
    if not colors:
        colors = [
            '#1f77b4',
            '#ff7f0e',
            '#2ca02c',
            '#d62728',
            '#9467bd',
            '#8c564b',
            '#e377c2',
            '#7f7f7f',
            '#bcbd22',
            '#17becf'
        ]
    return colors


def annotate_scan(scan, products, nperrow=4, ax=None, label=True):
    '''Given an MS1 :class:`~.ScanBase` ``scan`` and a :class:`~.Sequence` of
    :class:`~.ScanBase` product scans, draw the MS1 spectrum in full profile,
    and then in a subplot grid below it, draw a zoomed-in view of the MS1 spectrum
    surrounding the area around each precursor ion that gave rise to the scans in
    ``products``, with monoisotopic peaks and isolation windows marked.

    .. plot::
        :include-source:

        import ms_deisotope
        from ms_deisotope import plot
        from ms_deisotope.test.common import datafile

        reader = ms_deisotope.MSFileLoader(datafile("20150710_3um_AGP_001_29_30.mzML.gz"))
        bunch = next(reader)

        bunch.precursor.pick_peaks()
        bunch.precursor.deconvolute(
            scorer=ms_deisotope.PenalizedMSDeconVFitter(20., 2.0),
            averagine=ms_deisotope.glycopeptide, use_quick_charge=True)

        ax = plot.annotate_scan(bunch.precursor, bunch.products, nperrow=2)
        ax.figure.set_figwidth(12)
        ax.figure.set_figheight(16)


    Parameters
    ----------
    scan: ScanBase
        The precursor MS1 scan
    products: :class:`~.Sequence` of :class:`~.ScanBase`
        The collection of MSn scans based upon ``scan``
    nperrow: int
        The number of precursor ion subplots to draw per row
        of the grid. Defaults to :const:`4`.
    ax: :class:`matplotlib._axes.Axes`
        An :class:`~.Axes` object to use to find the figure to
        draw the plot on.

    Returns
    -------
    :class:`matplotlib._axes.Axes`:
        The axes of the full MS1 profile plot
    '''
    if ax is None:
        figure = plt.figure()
    else:
        figure = ax.figure
    n = len(products)
    if n > 0:
        gs = gridspec.GridSpec(1 + max(int(math.ceil(n / float(nperrow))), 1), nperrow)
    else:
        gs = gridspec.GridSpec(1 + (n / nperrow), nperrow)
    ax = figure.add_subplot(gs[0, :])
    if scan.is_profile:
        draw_raw(scan.arrays, ax=ax)
    if scan.peak_set is None:
        scan.pick_peaks()
    draw_peaklist(scan.peak_set, ax=ax, lw=0.5, alpha=0.75)
    if scan.deconvoluted_peak_set:
        draw_peaklist(scan.deconvoluted_peak_set, ax=ax, lw=0.5, alpha=0.75)
    ax.set_title(scan.id)
    k = -1
    # for the ith row of the grid
    for i in range(int(n // nperrow) + 1):
        # for the jth column of the row
        for j in range(nperrow):
            # the kth MS^n scan
            k += 1
            try:
                product_scan = products[k]
            except IndexError:
                # the number of MS^n scans is not a multiple of nperrow
                # so the last row has fewer than nperrow
                break
            # add the subplot from the grid spec using i + 1 instead of i
            # because the MS1 scan is drawn across the row at i = 0
            ax = figure.add_subplot(gs[i + 1, j])
            if scan.is_profile:
                draw_raw(scan.arrays, ax=ax, alpha=0.8)
            # obtain the interval around the precursor
            pinfo = product_scan.precursor_information
            if not product_scan.isolation_window.is_empty():
                lower, upper = (product_scan.isolation_window.lower_bound - 2,
                                product_scan.isolation_window.upper_bound + 2)
            else:
                lower = pinfo.mz - 4
                upper = pinfo.mz + 4
            try:
                peak = max(scan.peak_set.between(lower + 1.2, upper - 1.2), key=lambda x: x.intensity)
                local_intensity = peak.intensity
            except ValueError:
                if scan.deconvoluted_peak_set:
                    try:
                        peak = max(scan.deconvoluted_peak_set.between(lower + 1.2, upper - 1.2),
                                   key=lambda x: x.intensity, use_mz=True)
                        local_intensity = peak.intensity
                    except ValueError:
                        local_intensity = 1e3
                else:
                    local_intensity = 1e3

            # get the monoisotopic peak for the precursor, or the isolation center depending
            # upon whether the precursor has been deconvoluted and what the instrument reports
            if pinfo.extracted_charge != 0:
                target_mz = pinfo.extracted_mz
            else:
                target_mz = pinfo.mz

            draw_peaklist(scan.peak_set, ax=ax, alpha=0.5, lw=0.5)
            if scan.deconvoluted_peak_set:
                draw_peaklist(
                    scan.deconvoluted_peak_set.between(
                        lower - 1.2, upper + 1.2, use_mz=True),
                    ax=ax, alpha=0.9, color='blue')

            if label:
                label_peaks(scan, lower, upper, ax=ax,
                            is_deconvoluted=bool(scan.deconvoluted_peak_set))
            ax.set_ylim(0, local_intensity * 1.25)
            ax.set_xlim(lower, upper)
            upper_intensity = local_intensity

            # draw the precursor isolation annotations
            ax.vlines(target_mz, 0, upper_intensity * 1.5, alpha=0.50,
                      color='red', lw=1)
            if product_scan.isolation_window.lower != 0:
                ax.vlines(product_scan.isolation_window.lower_bound, 0,
                          upper_intensity * 1.5, linestyle='--', alpha=0.5)
            if product_scan.isolation_window.upper != 0:
                ax.vlines(product_scan.isolation_window.upper_bound, 0,
                          upper_intensity * 1.5, linestyle='--', alpha=0.5)
            ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            ax.set_ylabel("")
            ax.set_xlabel("")
            if pinfo.extracted_charge == 0:
                if pinfo.charge == "ChargeNotProvided":
                    charge = '?'
                else:
                    charge = str(pinfo.charge)
            else:
                charge = str(pinfo.extracted_charge)
            ax.set_title("%0.3f @ %s" % (target_mz, charge))
    fig = figure
    # this should probably be a function of figwidth and number of rows
    fig.set_figheight(fig.get_figheight() * 2)
    fig.tight_layout()
    return ax


def annotate_scan_single(scan, product_scan, ax=None, label=True, standalone=True):
    '''Draw a zoomed-in view of the MS1 spectrum ``scan`` surrounding the
    area around each precursor ion that gave rise to ``product_scan``
    with monoisotopic peaks and isolation windows marked.

    .. plot::
        :include-source:

        import ms_deisotope
        from ms_deisotope import plot
        from ms_deisotope.test.common import datafile

        reader = ms_deisotope.MSFileLoader(datafile("20150710_3um_AGP_001_29_30.mzML.gz"))
        bunch = next(reader)

        bunch.precursor.pick_peaks()
        bunch.precursor.deconvolute(
            scorer=ms_deisotope.PenalizedMSDeconVFitter(20., 2.0),
            averagine=ms_deisotope.glycopeptide, use_quick_charge=True)

        ax = plot.annotate_scan_single(bunch.precursor, bunch.products[0])
        ax.figure.set_figwidth(12)



    Parameters
    ----------
    scan: ScanBase
        The MS1 scan to annotate
    product_scan: ScanBase
        The product scan to annotate the precursor ion of
    ax: :class:`matplotlib._axes.Axes`
        An :class:`~.Axes` object to draw the plot on

    Returns
    -------
    :class:`matplotlib._axes.Axes`
    '''
    if ax is None:
        _, ax = plt.subplots(1)

    if scan.is_profile:
        draw_raw(scan.arrays, ax=ax)
    if scan.peak_set is None:
        scan.pick_peaks()
    draw_peaklist(scan.peak_set, ax=ax, lw=0.5, alpha=0.75)
    if scan.deconvoluted_peak_set:
        draw_peaklist(scan.deconvoluted_peak_set, ax=ax, lw=0.5, alpha=0.75)

    pinfo = product_scan.precursor_information
    if not product_scan.isolation_window.is_empty():
        lower, upper = (product_scan.isolation_window.lower_bound - 2,
                        product_scan.isolation_window.upper_bound + 2)
    else:
        lower = pinfo.mz - 4
        upper = pinfo.mz + 4

    peak_set = scan.peak_set
    try:
        peak = max(peak_set.between(lower + 1.2, upper - 1.2), key=lambda x: x.intensity)
        local_intensity = peak.intensity
    except ValueError:
        if scan.deconvoluted_peak_set:
            try:
                peak = max(scan.deconvoluted_peak_set.between(lower + 1.2, upper - 1.2),
                           key=lambda x: x.intensity, use_mz=True)
                local_intensity = peak.intensity
            except ValueError:
                local_intensity = 1e3
        else:
            local_intensity = 1e3

    # get the monoisotopic peak for the precursor, or the isolation center depending
    # upon whether the precursor has been deconvoluted and what the instrument reports
    if pinfo.extracted_charge != 0:
        target_mz = pinfo.extracted_mz
    else:
        target_mz = pinfo.mz

    draw_peaklist(scan.peak_set, ax=ax, alpha=0.5, lw=0.5)
    if scan.deconvoluted_peak_set:
        draw_peaklist(
            scan.deconvoluted_peak_set.between(
                lower - 1.2, upper + 1.2, use_mz=True),
            ax=ax, alpha=0.9, color='blue')

    if label:
        label_peaks(scan, lower, upper, ax=ax,
                    is_deconvoluted=bool(scan.deconvoluted_peak_set))

    ax.set_ylim(0, local_intensity * 1.25)
    ax.set_xlim(lower, upper)
    upper_intensity = local_intensity

    # draw the precursor isolation annotations
    ax.vlines(target_mz, 0, upper_intensity * 1.5, alpha=0.50,
              color='red', lw=1)
    if product_scan.isolation_window.lower != 0:
        ax.vlines(product_scan.isolation_window.lower_bound, 0,
                  upper_intensity * 1.5, linestyle='--', alpha=0.5)
    if product_scan.isolation_window.upper != 0:
        ax.vlines(product_scan.isolation_window.upper_bound, 0,
                  upper_intensity * 1.5, linestyle='--', alpha=0.5)
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    ax.set_ylabel("")
    ax.set_xlabel("")
    if pinfo.extracted_charge == 0:
        if pinfo.charge == "ChargeNotProvided":
            charge = '?'
        else:
            charge = str(pinfo.charge)
    else:
        charge = str(pinfo.extracted_charge)
    ax.set_title("%0.3f @ %s" % (target_mz, charge))

    return ax


def annotate_isotopic_peaks(scan, ax=None, color_cycle=None, **kwargs):
    '''Mark distinct isotopic peaks from the :class:`~.DeconvolutedPeakSet`
    in ``scan``.

    .. plot::
        :include-source:

        import ms_deisotope
        from ms_deisotope import plot
        from ms_deisotope.test.common import datafile

        reader = ms_deisotope.MSFileLoader(datafile("20150710_3um_AGP_001_29_30.mzML.gz"))
        bunch = next(reader)

        bunch.precursor.pick_peaks()
        bunch.precursor.deconvolute(
            scorer=ms_deisotope.PenalizedMSDeconVFitter(20., 2.0),
            averagine=ms_deisotope.glycopeptide, use_quick_charge=True)

        ax = plot.draw_peaklist(bunch.precursor, color='black')
        ax = plot.annotate_isotopic_peaks(bunch.precursor, ax=ax)
        ax.set_xlim(1160, 1165)
        ax.figure.set_figwidth(12)


    Parameters
    ----------
    scan: ScanBase
        The scan to annotate
    color_cycle: :class:`~.Iterable`
        An iterable to draw isotopic cluster colors from
    ax: :class:`matplotlib._axes.Axes`
        An :class:`~.Axes` object to draw the plot on

    Returns
    -------
    :class:`matplotlib._axes.Axes`
    '''
    from .peak_set import DeconvolutedPeakSet

    if ax is None:
        _, ax = plt.subplots(1)
    if color_cycle is None:
        color_cycle = _default_color_cycle()
    color_cycle = itertools.cycle(color_cycle)
    if isinstance(scan, DeconvolutedPeakSet):
        peaks = scan
    else:
        peaks = getattr(scan, "deconvoluted_peak_set", [])
    for peak in peaks:
        color = next(color_cycle)
        draw_peaklist(peak.envelope, ax=ax, color=color, alpha=0.75, **kwargs)
        ax.scatter(*zip(*peak.envelope), color=color, alpha=0.75)
    return ax


def label_peaks(scan, min_mz=None, max_mz=None, ax=None, is_deconvoluted=None, threshold=None, **kwargs):
    """Label a region of the peak list, marking centroids with their m/z or mass. If the peaks
    of `scan` have been deconvoluted, the most abundant peak will be annotated with
    "<neutral mass> (<charge>)", otherwise just "<m/z>".

    Parameters
    ----------
    scan : :class:`~.ScanBase`
        The scan to annotate
    min_mz : float, optional
        The minimum m/z to annotate
    max_mz : float, optional
        The maximum m/z to annotate
    ax: :class:`matplotlib._axes.Axes`
        An :class:`~.Axes` object to draw the plot on
    is_deconvoluted : bool, optional
        Whether or not to always use :attr:`Scan.deconvoluted_peak_set`
    threshold : float, optional
        The intensity threshold under which peaks will be ignored

    Returns
    -------
    ax: :class:`matplotlib._axes.Axes`
        The axes the plot was drawn on
    annotations: :class:`list` of :class:`matplotlib.text.Text`
        The list of :class:`matplotlib.text.Text` annotations
    """
    if ax is None:
        # when no axes are provided, draw all the peak data
        _, ax = plt.subplots(1)
        if scan.is_profile:
            draw_raw(scan, ax)
        if scan.peak_set is not None:
            draw_peaklist(scan.peak_set, ax)
        if scan.deconvoluted_peak_set is not None:
            draw_peaklist(scan.deconvoluted_peak_set, ax)
    if is_deconvoluted is None:
        is_deconvoluted = scan.deconvoluted_peak_set is not None
    if min_mz is None:
        min_mz = 0
    if max_mz is None:
        if is_deconvoluted:
            max_mz = max(peak.mz for peak in scan.deconvoluted_peak_set)
        else:
            max_mz = max(peak.mz for peak in scan.peak_set)
    annotations = []
    # select the peak sub range
    if is_deconvoluted:
        subset = scan.deconvoluted_peak_set.between(
            min_mz, max_mz, use_mz=True)
    else:
        subset = scan.peak_set.between(
            min_mz, max_mz)
    if not subset:
        return
    # guess the threshold
    if threshold is None:
        threshold = 0.0
        if is_deconvoluted:
            threshold_list = ([max(i.intensity for i in p.envelope)
                               for p in subset])
        else:
            threshold_list = ([p.intensity for p in subset])
        if threshold_list:
            threshold = np.mean(threshold_list)
        threshold_list = [v > threshold for v in threshold_list]
        if threshold_list:
            threshold = np.mean(threshold_list)
    # draw the actual labels
    kwargs.setdefault("clip_on", True)
    kwargs.setdefault("fontsize", 10)
    if is_deconvoluted:
        for peak in subset:
            if peak.intensity > threshold:
                label = '%0.2f (%d)' % (peak.neutral_mass, peak.charge)
                # set the y-position to the highest peak in the isotopic
                # pattern
                pt = max(peak.envelope, key=lambda x: x.intensity)
                y = pt.intensity * 1.05
                # set the x-position to the weighted average m/z in the
                # isotopic pattern
                x = np.average(
                    [p.mz for p in peak.envelope],
                    weights=[p.intensity for p in peak.envelope])
                annotations.append(
                    ax.text(x, y, label, ha='center', **kwargs))
    else:
        for peak in subset:
            if peak.intensity > threshold:
                label = "%0.2f" % (peak.mz, )
                y = peak.intensity * 1.05
                x = peak.mz
                annotations.append(
                    ax.text(x, y, label, ha='center', **kwargs))
    return annotations, ax


def draw_spectrum_paths(graph, edge_color='red', peak_color='orange', alpha=0.8, fontsize=12, ax=None, **kwargs):
    """Draw all the paths given by `graph` over a rendered peak list.

    This function will draw all peaks which an edge connects in `peak_color`, and
    draws edges as lines connecting peaks in `edge_color`, with their annotations
    written across them.

    If a peak is not connected to an edge, it will not be drawn.

    Parameters
    ----------
    graph : :class:`~.SpectrumGraph` or :class:`list` of :class:`~.spectrum_graph.Path`
        The paths to annotate. If a :class:`~.SpectrumGraph` is given, the top 100 longest
        paths will be enumerated from it.
    edge_color : str, optional
        The color to draw the edge lines in (the default is 'red')
    peak_color : str, optional
        The color to draw the connected peaks in (the default is 'orange')
    alpha : float, optional
        The alpha channel value for peaks (the default is 0.8)
    fontsize : int, optional
        The font size to use for edge annotations (the default is 12)
    ax : :class:`matplotlib._axes.Axes`, optional
        The axes to draw the plot on (the default is None, which will cause a new figure with a
        single axes to be created)

    Returns
    -------
    :class:`matplotlib._axes.Axes`
    """
    if ax is None:
        _, ax = plt.subplots(1)
    try:
        paths = graph.longest_paths(limit=100)
    except AttributeError:
        paths = list(graph)

    for path in paths:
        for edge in path:
            for p1 in edge.start:
                for p2 in edge.end:
                    _draw_peak_pair(
                        (p1, p2), edge_color, peak_color, alpha, fontsize,
                        label=edge.annotation, ax=ax, **kwargs)


def _draw_peak_pair(pair, edge_color='red', peak_color='orange', alpha=0.8, fontsize=12, label=None, rotation=45,
                    ax=None, **kwargs):
    p1, p2 = pair
    ax.plot((p1.mz, p2.mz), (p1.intensity, p2.intensity),
            color=edge_color, alpha=alpha, **kwargs)
    kwargs.setdefault("clip_on", False)
    clip_on = kwargs['clip_on']
    draw_peaklist(pair, ax=ax, alpha=0.4, color=peak_color)
    if label:
        midx = (p1.mz + p2.mz) / 2
        # interpolate the midpoint's height
        midy = (p1.intensity * (p2.mz - midx) +
                p2.intensity * (midx - p1.mz)) / (p2.mz - p1.mz)

        # find the angle of the line connecting the two peaks
        xlo = min(p1.mz, p2.mz)
        xhi = max(p1.mz, p2.mz)
        adj = xhi - xlo
        ylo = min(p1.intensity, p2.intensity)
        yhi = max(p1.intensity, p2.intensity)
        opp = yhi - ylo
        hypot = np.hypot(adj, opp) # pylint: disable=assignment-from-no-return
        rotation = np.arccos(adj / hypot)  # pylint: disable=assignment-from-no-return

        if isinstance(label, (list, tuple)):
            label = '-'.join(map(str, label))
        else:
            label = str(label)
        ax.text(midx, midy, label, fontsize=fontsize,
                ha='center', va='bottom', rotation=rotation, clip_on=clip_on)



import unittest

from ms_deisotope.test.common import datafile, gzload

from ms_deisotope.data_source import common, mzml
from ms_deisotope.averagine import peptide
from ms_deisotope.scoring import MSDeconVFitter


class TestIntegration(unittest.TestCase):
    def get_scan(self):
        scan_data = gzload(datafile("test_scan.pkl.gz"))
        scan = common.Scan(scan_data, mzml.MzMLDataInterface())
        return scan

    def get_reference(self):
        processed_scan = gzload(datafile("test_scan_results.pkl.gz"))
        return processed_scan

    def test_scan(self):
        scan = self.get_scan()
        scan.pick_peaks()
        scan.deconvolute(averagine=peptide, scorer=MSDeconVFitter(10.0),
                         ignore_below=0.05, truncate_after=0.8, charge_range=(1, 4))
        reference = self.get_reference()
        for peak in scan.deconvoluted_peak_set:
            peaks = reference.deconvoluted_peak_set.all_peaks_for(peak.neutral_mass, 1e-6)
            for ref in peaks:
                if ref.charge == peak.charge:
                    self.assertTrue(
                        abs(peak.neutral_mass - ref.neutral_mass) < 1e-2,
                        peak.neutral_mass - ref.neutral_mass)
                    self.assertTrue(
                        abs(peak.intensity - ref.intensity) < 1e-2,
                        peak.intensity - ref.intensity)
                    self.assertTrue(
                        abs(peak.score - ref.score) < 1e-2,
                        peak.score - ref.score)
                    for at, bt in zip(peak.envelope, ref.envelope):
                        self.assertTrue(
                            abs(at.mz - bt.mz) < 1e-2,
                            at.mz - bt.mz)
                        self.assertTrue(
                            abs(at.intensity - bt.intensity) < 1e-2,
                            at.intensity - bt.intensity)


if __name__ == '__main__':
    unittest.main()



import multiprocessing
import os
import re
import sys
import warnings
import logging

import click

from brainpy import periodic_table

from ms_deisotope.averagine import (
    Averagine, glycan as n_glycan_averagine, permethylated_glycan,
    peptide, glycopeptide, heparin, heparan_sulfate)


def processes_option(f):
    opt = click.option(
        "-p", "--processes", 'processes', type=click.IntRange(1, multiprocessing.cpu_count()),
        default=min(multiprocessing.cpu_count(), 4), help=('Number of worker processes to use. Defaults to 4 '
                                                           'or the number of CPUs, whichever is lower'))
    return opt(f)


def validate_element(element):
    valid = element in periodic_table
    if not valid:
        raise click.Abort("%r is not a valid element" % element)
    return valid


def parse_averagine_formula(formula):
    if isinstance(formula, Averagine):
        return formula
    return Averagine({k: float(v) for k, v in re.findall(r"([A-Z][a-z]*)([0-9\.]*)", formula)
                      if float(v or 0) > 0 and validate_element(k)})


averagines = {
    'glycan': n_glycan_averagine,
    'permethylated-glycan': permethylated_glycan,
    'peptide': peptide,
    'glycopeptide': glycopeptide,
    'heparin': heparin,
    "heparan-sulfate": heparan_sulfate
}


def validate_averagine(averagine_string):
    if isinstance(averagine_string, Averagine):
        return averagine_string
    if averagine_string in averagines:
        return averagines[averagine_string]
    else:
        return parse_averagine_formula(averagine_string)


class AveragineParamType(click.types.StringParamType):
    name = "MODEL"

    models = averagines

    def convert(self, value, param, ctx):
        return validate_averagine(value)

    def get_metavar(self, param):
        return '[%s]' % '|'.join(sorted(averagines.keys()))

    def get_missing_message(self, param):
        return 'Choose from %s, or provide a formula.' % ', '.join(self.choices)


def register_debug_hook():
    import traceback

    def info(type, value, tb):
        if hasattr(sys, 'ps1') or not sys.stderr.isatty():
            sys.__excepthook__(type, value, tb)
        else:
            try:
                import ipdb as pdb_api
            except ImportError:
                import pdb as pdb_api
            traceback.print_exception(type, value, tb)
            pdb_api.post_mortem(tb)

    sys.excepthook = info
    logging.basicConfig(level="DEBUG")


def is_debug_mode():
    env_val = os.environ.get('MS_DEISOTOPE_DEBUG', '').lower()
    if not env_val:
        return False
    if env_val in ('0', 'no', 'false', 'off'):
        return False
    elif env_val in ('1', 'yes', 'true', 'on'):
        return True
    else:
        warnings.warn("MS_DEISOTOPE_DEBUG value %r was not recognized. Enabling debug mode" % (env_val, ))
        return True



from .infer_type import MSFileLoader
from .mzml import MzMLLoader
from .mzxml import MzXMLLoader
from .mgf import MGFLoader
from .common import (
    Scan,
    PrecursorInformation, ProcessedScan,
    IsolationWindow,
    ScanDataSource, ScanIterator, ScanBunch,
    RandomAccessScanSource, ChargeNotProvided)

from .metadata.file_information import (
    FileInformation,
    FileFormat, file_formats,
    FileContent, content_keys,
    IDFormat, id_formats)

from .metadata.instrument_components import (
    Component, ComponentGroup, InstrumentInformation,
    components)

from .metadata.scan_traits import (
    ScanAcquisitionInformation, ScanEventInformation,
    ScanWindow, scan_attributes)

from .metadata.activation import (
    ActivationInformation, MultipleActivationInformation,
    DissociationMethod, dissociation_methods)

from .scan import ScanProxyContext

from ._compression import get_opener

__all__ = [
    "MSFileLoader", "MzMLLoader",
    "MzXMLLoader", "MGFLoader",

    "Scan", "ProcessedScan",

    "PrecursorInformation",

    "ActivationInformation", "MultipleActivationInformation",
    "DissociationMethod", "dissociation_methods",

    "ScanAcquisitionInformation", "ScanEventInformation",
    "IsolationWindow", "ScanWindow", "scan_attributes",

    "ScanDataSource", "ScanIterator", "ScanBunch",
    "ScanWindow", "RandomAccessScanSource", "ChargeNotProvided",
    "get_opener", "ScanProxyContext",
]



from collections import namedtuple, MutableSequence

from .cv import Term, TermSet


def _isclose(x, y, atol=1e-3):
    return abs(x - y) < atol


class IsolationWindow(namedtuple("IsolationWindow", ['lower', 'target', 'upper'])):
    r"""Describes the m/z interval a precursor ion was isolated from in the precursor scan

    An :class:`IsolationWindow` instance is comparable, hashable, and orderable. It is based
    on a :class:`namedtuple`, and supports the same interface and methods as a :class:`tuple`.

    Attributes
    ----------
    lower: float
        The distance from the center of the isolation window towards 0
    target: float
        The center of the isolation window in m/z
    upper: float
        The distance from the center of the isolation window towards :math:`\infty`
    lower_bound: float
        The m/z coordinate of the lower bound :attr:`target` - :attr:`lower`
    upper_bound: float
        The m/z coordinate of the upper bound :attr:`target` + :attr:`upper`
    width: float
        The sum of :attr:`lower` and :attr:`upper`, the total m/z space spanned by
        the window
    """

    @classmethod
    def make_empty(cls, point):
        """A helper method for instantiating an empty isolation window centered at
        `point`

        Parameters
        ----------
        point : float
            The point to center the new empty window at.

        Returns
        -------
        :class:`IsolationWindow`
        """
        return cls(0, point, 0)

    @property
    def lower_bound(self):
        """The m/z coordinate of the lower bound :attr:`target` - :attr:`lower`
        """
        return self.target - self.lower

    @property
    def upper_bound(self):
        """The m/z coordinate of the upper bound :attr:`target` + :attr:`upper`
        """
        return self.target + self.upper

    @property
    def width(self):
        '''The sum of :attr:`lower` and :attr:`upper`, the total m/z space spanned by
        the window'''
        return self.lower + self.upper

    def __contains__(self, x):
        return self.spans(x, 0.1)

    def spans(self, x, tolerance=0.1):
        """Test whether `x` is within the interval defined by this window's bounds.

        This method permits a small amount of error (controlled by `tolerance`) when
        computing the bounds.

        Equivalent to: :code:`(self.lower_bound - tolerance) <= x <= (self.upper_bound + tolerance)`

        Parameters
        ----------
        x : float
            The number to query
        tolerance : float, optional
            The amount of error to accept when computing the bounds (the default is 0.1)

        Returns
        -------
        :class:`bool`
        """
        return (self.lower_bound - tolerance) <= x <= (self.upper_bound + tolerance)

    def is_empty(self):
        """Tests if the window is empty (e.g. its upper bound is equal to its lower bound)

        Returns
        -------
        :class:`bool`
        """
        if self.lower is None:
            return self.upper is None
        return self.lower == self.upper == 0.0

    def __nonzero__(self):
        return not self.is_empty()

    def __bool__(self):
        return self.__nonzero__()

    def __eq__(self, other):
        return _isclose(self.lower, other.lower) and _isclose(
            self.upper, other.upper) and _isclose(
                self.target, other.target)

    def __ne__(self, other):
        return not (self == other)


class ScanAcquisitionInformation(MutableSequence):
    """Describes the set distinct scans along the measurable range by
    the instrument that were performed to produce the acquired data

    :class:`ScanAcquisitionInformation` objects implement the :class:`MutableSequence`
    interface, acting as a sequence of :class:`ScanEventInformation` objects.

    Attributes
    ----------
    combination : :class:`str`
        A controlled vocabulary string describing the way in which scans were combined
    scan_list : :class:`list` of :class:`ScanEventInformation`
        The list of scan events performed
    """

    def __init__(self, combination, scan_list):  # pylint: disable=super-init-not-called
        self.combination = combination
        self.scan_list = scan_list

    def __getitem__(self, i):
        return self.scan_list[i]

    def __setitem__(self, i, value):
        self.scan_list[i] = value

    def __delitem__(self, i):
        del self.scan_list[i]

    def insert(self, index, value):
        """Insert an event at a given position

        Parameters
        ----------
        index : int
            The position to add the new event
        value : :class:`ScanEventInformation`
            The event to insert
        """
        if not isinstance(value, ScanEventInformation):
            raise TypeError("Expected ScanEventInformation but got %r" % (type(value), ))
        self.scan_list.insert(index, value)

    def __len__(self):
        return len(self.scan_list)

    def __repr__(self):
        return "ScanAcquisitionInformation(combination=%r, scan_list=%r)" % (
            self.combination, self.scan_list)

    def __eq__(self, other):
        return self.combination == other.combination and self.scan_list == other.scan_list

    def __ne__(self, other):
        return not (self == other)


class ScanEventInformation(object):
    def __init__(self, start_time, window_list, drift_time=None, injection_time=None, traits=None, **options):
        self.start_time = start_time
        self.window_list = window_list or []
        self.drift_time = drift_time
        self.injection_time = injection_time
        self.traits = traits or {}
        self.traits.update(options)

    def has_ion_mobility(self):
        return self.drift_time is not None and self.drift_time > 0

    def __getitem__(self, i):
        return self.window_list[i]

    def __iter__(self):
        return iter(self.window_list)

    def __len__(self):
        return len(self.window_list)

    def __repr__(self):
        template = "ScanEventInformation(start_time={}, window_list={}{})"
        if self.has_ion_mobility():
            tail = ", drift_time={}".format(self.drift_time)
        else:
            tail = ''
        if self.injection_time is not None:
            tail += ', injection_time={}'.format(self.injection_time)
        if self.traits:
            tail += ", traits={}".format(self.traits)
        form = template.format(self.start_time, self.window_list, tail)
        return form

    def __eq__(self, other):
        eq = _isclose(self.start_time, other.start_time) and (
            self.window_list == other.window_list)
        if not eq:
            return False
        if self.has_ion_mobility() != other.has_ion_mobility():
            return False
        if self.has_ion_mobility() and not _isclose(self.drift_time, other.drift_time):
            return False
        return True

    def __ne__(self, other):
        return not (self == other)

    def total_scan_window(self):
        low = float('inf')
        high = 0
        for window in self:
            low = min(low, window.lower)
            high = max(high, window.upper)
        return ScanWindow(low, high)


class ScanWindow(namedtuple("ScanWindow", ['lower', 'upper'])):
    def __contains__(self, i):
        return self.lower <= i <= self.upper

    def is_empty(self):
        if self.lower is None:
            return self.upper is None
        return self.lower == self.upper == 0.0

    def __nonzero__(self):
        return not self.is_empty()

    def __bool__(self):
        return self.__nonzero__()

    def __eq__(self, other):
        return _isclose(self.lower, other.lower) and _isclose(
            self.upper, other.upper)

    def __ne__(self, other):
        return not (self == other)


class ScanAttribute(Term):
    pass


scan_attributes = []

# [[[cog
# import cog
# from ms_deisotope.data_source.metadata.cv import render_list
# render_list('scan attribute', term_cls_name="ScanAttribute", writer=cog.out)
# ]]]
scan_attributes = TermSet([
    ScanAttribute(u'ion mobility attribute', u'MS:1002892',
                  (u'An attribute describing ion mobility searches.'),
                  'scan attribute',
                  [u'scan attribute', u'PSM-level attribute', u'object attribute', u'single identification result attribute', u'identification attribute', u'analysis attribute', u'spectrum identification result details']),
    ScanAttribute(u'mass resolution', u'MS:1000011',
                  (u'Smallest mass difference between two equal magnitude peaks'
                   u'so that the valley between them is a specified fraction of'
                   u'the peak height.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'scan start time', u'MS:1000016',
                  (u'The time that an analyzer started a scan, relative to the'
                   u'start of the MS run.'),
                  'scan attribute',
                  [u'scan attribute', u'PSM-level attribute', u'object attribute', u'single identification result attribute', u'identification attribute', u'analysis attribute', u'spectrum identification result details']),
    ScanAttribute(u'scan rate', u'MS:1000015',
                  (u'Rate in Th/sec for scanning analyzers.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'elution time', u'MS:1000826',
                  (u'The time of elution from all used chromatographic columns'
                   u'(one or more) in the chromatographic separation step,'
                   u'relative to the start of the chromatography.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'mass resolving power', u'MS:1000800',
                  (u'The observed mass divided by the difference between two'
                   u'masses that can be separated: m/dm. The procedure by which'
                   u'dm was obtained and the mass at which the measurement was'
                   u'made should be reported.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'analyzer scan offset', u'MS:1000803',
                  (u'Offset between two analyzers in a constant neutral loss or'
                   u'neutral gain scan. The value corresponds to the neutral loss'
                   u'or neutral gain value.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'ion injection time', u'MS:1000927',
                  (u'The length of time spent filling an ion trapping device.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'instrument specific scan attribute', u'MS:1002527',
                  (u'Instrument specific scan properties that are associated with'
                   u'a value.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'interchannel delay', u'MS:1000880',
                  (u'The duration of intervals between scanning, during which the'
                   u'instrument configuration is switched.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'dwell time', u'MS:1000502',
                  (u'The time spent gathering data across a peak.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'filter string', u'MS:1000512',
                  (u'A string unique to Thermo instrument describing instrument'
                   u'settings for the scan.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'preset scan configuration', u'MS:1000616',
                  (u'A user-defined scan configuration that specifies the'
                   u'instrumental settings in which a spectrum is acquired. An'
                   u'instrument may cycle through a list of preset scan'
                   u'configurations to acquire data. This is a more generic term'
                   u'for the Thermo \\"scan event\\", which is defined in the'
                   u'Thermo Xcalibur glossary as: a mass spectrometer scan that'
                   u'is defined by choosing the necessary scan parameter'
                   u'settings. Multiple scan events can be defined for each'
                   u'segment of time.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'second column elution time', u'MS:1002083',
                  (u'The time of elution from the second chromatographic column'
                   u'in the chromatographic separation step, relative to the'
                   u'start of the chromatography on the second column.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'first column elution time', u'MS:1002082',
                  (u'The time of elution from the first chromatographic column in'
                   u'the chromatographic separation step, relative to the start'
                   u'of chromatography on the first column.'),
                  'scan attribute',
                  [u'scan attribute', u'object attribute']),
    ScanAttribute(u'inverse reduced ion mobility', u'MS:1002815',
                  (u'Ion mobility measurement for an ion or spectrum of ions as'
                   u'measured in an ion mobility mass spectrometer. This might'
                   u'refer to the central value of a bin into which all ions'
                   u'within a narrow range of mobilities have been aggregated.'),
                  'scan attribute',
                  [u'ion selection attribute', u'ion mobility attribute', u'object attribute', u'scan attribute', u'PSM-level attribute', u'single identification result attribute', u'identification attribute', u'analysis attribute', u'spectrum identification result details']),
    ScanAttribute(u'ion mobility drift time', u'MS:1002476',
                  (u'Drift time of an ion or spectrum of ions as measured in an'
                   u'ion mobility mass spectrometer. This time might refer to the'
                   u'central value of a bin into which all ions within a narrow'
                   u'range of drift time have been aggregated.'),
                  'scan attribute',
                  [u'ion selection attribute', u'ion mobility attribute', u'object attribute', u'scan attribute', u'PSM-level attribute', u'single identification result attribute', u'identification attribute', u'analysis attribute', u'spectrum identification result details']),
    ScanAttribute(u'FAIMS compensation voltage', u'MS:1001581',
                  (u'The DC potential applied to the asymmetric waveform in FAIMS'
                   u'that compensates for the difference between high and low'
                   u'field mobility of an ion.'),
                  'scan attribute',
                  [u'ion mobility attribute', u'scan attribute', u'PSM-level attribute', u'object attribute', u'single identification result attribute', u'identification attribute', u'analysis attribute', u'spectrum identification result details']),
    ScanAttribute(u'synchronous prefilter selection', u'MS:1002528',
                  (u'Synchronous prefilter selection.'),
                  'scan attribute',
                  [u'instrument specific scan attribute', u'scan attribute', u'object attribute']),
])
# [[[end]]]

__all__ = [
    "IsolationWindow", "ScanAcquisitionInformation", "ScanEventInformation",
    "ScanWindow", "ScanAttribute", "scan_attributes"
]



import unittest

import numpy as np


def make_peak_set_test_suite(peak_set_cls, peak_cls):

    class TestDeconvolutedPeakSet(unittest.TestCase):
        def test_search(self):
            x = np.arange(1000, 1200, 0.001)
            y = np.ones_like(x)

            peaks = [peak_cls(x[i], y[i], 1, 1, None, 0) for i in range(len(x))]
            ps = peak_set_cls(peaks)
            ps.reindex()
            t = ps[5000].neutral_mass
            assert ps.has_peak(t) is ps[5000]

            for peak in ps.all_peaks_for(t):
                assert abs((peak.neutral_mass - t) / t) < 1e-5

        def test_missing_interval(self):
            x = np.arange(1000, 1200, 0.5)
            y = np.ones_like(x)

            peaks = [peak_cls(x[i], y[i], 1, 1, None, 0) for i in range(len(x))]
            ps = peak_set_cls(peaks)
            ps.reindex()

            for xi in np.arange(1000, 1200, 0.01):
                for p in ps.all_peaks_for(xi):
                    assert abs((xi - p.neutral_mass) / p.neutral_mass) < 1e-5

    return TestDeconvolutedPeakSet


try:
    from ms_deisotope._c.peak_set import (
        DeconvolutedPeakSet as CDeconvolutedPeakSet,
        DeconvolutedPeak as CDeconvolutedPeak,
        DeconvolutedPeakSetIndexed as CDeconvolutedPeakSetIndexed)

    TestCDeconvolutedPeakSet = make_peak_set_test_suite(CDeconvolutedPeakSet, CDeconvolutedPeak)
    TestCDeconvolutedPeakSetIndexed = make_peak_set_test_suite(CDeconvolutedPeakSetIndexed, CDeconvolutedPeak)

    from ms_deisotope.peak_set import (
        _DeconvolutedPeak, _DeconvolutedPeakSet)

    TestPythonDeconvolutedPeakSet = make_peak_set_test_suite(_DeconvolutedPeakSet, _DeconvolutedPeak)

except ImportError:
    from ms_deisotope.peak_set import (
        _DeconvolutedPeak, _DeconvolutedPeakSet)

    TestPythonDeconvolutedPeakSet = make_peak_set_test_suite(_DeconvolutedPeakSet, _DeconvolutedPeak)


if __name__ == '__main__':
    unittest.main()



import unittest

from ms_deisotope.output import ProcessedMzMLDeserializer
from ms_deisotope.data_source import get_opener

from ms_deisotope.clustering import scan_clustering

from ms_deisotope.test.common import datafile


class TestScanClustering(unittest.TestCase):
    path = datafile("AGP_tryptic_300ng_2microscans_glycoproteomics_nCE_27-30.preprocessed.mzML.gz")

    @property
    def reader(self):
        reader = ProcessedMzMLDeserializer(get_opener(self.path))
        return reader

    def load_msms_scans(self, reader):
        products = list(map(reader.get_scan_by_id, reader.extended_index.msn_ids.keys()))
        return products

    def cluster_scans(self, scans):
        clusters = scan_clustering.cluster_scans(scans)
        return clusters

    def test_cluster_scans(self):
        reader = self.reader
        scans = self.load_msms_scans(reader)
        clusters = self.cluster_scans(scans)
        assert len(clusters) == 1124


if __name__ == "__main__":
    unittest.main()



from collections import defaultdict
from .lcms_feature import LCMSFeature
from ms_deisotope.data_source.common import ProcessedScan
from ms_deisotope import DeconvolutedPeakSet
from ms_peak_picker import PeakSet


class LCMSFeatureMap(object):

    def __init__(self, features):
        self.features = sorted(features, key=lambda x: x.mz)

    def __len__(self):
        return len(self.features)

    def __iter__(self):
        return iter(self.features)

    def __getitem__(self, i):
        if isinstance(i, (int, slice)):
            return self.features[i]
        else:
            return [self.features[j] for j in i]

    def search(self, mz, error_tolerance=2e-5):
        i = binary_search(self.features, mz, error_tolerance)
        if i is None:
            return None
        match = self[i]
        return match

    def find_all(self, mz, error_tolerance=2e-5):
        bounds = search_sweep(self.features, mz, error_tolerance)
        if bounds is not None:
            lo, hi = bounds
            return self[lo:hi]
        else:
            return []

    def index_range(self, lo, hi, error_tolerance=2e-5):
        return (
            binary_search_with_flag(
                self.features, lo, error_tolerance)[0][0],
            binary_search_with_flag(
                self.features, hi, error_tolerance)[0][0])

    def between(self, lo, hi, error_tolerance=2e-5):
        lo_ix = binary_search_with_flag(
            self.features, lo, error_tolerance)[0][0]
        hi_ix = binary_search_with_flag(
            self.features, hi, error_tolerance)[0][0]
        if self[lo_ix].mz < lo:
            lo_ix += 1
        if self[hi_ix] > hi:
            hi_ix -= 1
        if hi_ix < 0:
            hi_ix = 0
        if lo_ix > len(self):
            lo_ix = len(self) - 1
        return self[lo_ix:hi_ix]

    def __repr__(self):
        return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))


try:
    from ms_deisotope._c.feature_map.feature_map import LCMSFeatureMap
except ImportError:
    pass


class LCMSFeatureForest(LCMSFeatureMap):
    """An an algorithm for aggregating features from peaks of close mass
    weighted by intensity.

    This algorithm assumes that mass accuracy is correlated with intensity, so
    the most intense peaks should most accurately reflect their true neutral mass.
    The expected input is a list of (scan id, peak) pairs. This list is sorted by
    descending peak intensity. For each pair, using binary search, locate the nearest
    existing feature in :attr:`features`. If the nearest feature is within
    :attr:`error_tolerance` ppm of the peak's neutral mass, add this peak to that
    feature, otherwise create a new feature containing this peak and insert
    it into :attr:`features` while preserving the overall sortedness. This algorithm
    is carried out by :meth:`aggregate_unmatched_peaks`

    This process may produce features with large gaps in them, which
    may or may not be acceptable. To break gapped features into separate
    entities, the :class:`LCMSFeatureFilter` type has a method :meth:`split_sparse`.

    Attributes
    ----------
    features : list of LCMSFeature
        A list of growing LCMSFeature objects, ordered by neutral mass
    count : int
        The number of peaks accumulated
    error_tolerance : float
        The mass error tolerance between peaks and possible features (in ppm)
    """
    def __init__(self, features=None, error_tolerance=1e-5):
        if features is None:
            features = []
        self.features = sorted(features, key=lambda x: x.mz)
        self.error_tolerance = error_tolerance
        self.count = 0

    def find_insertion_point(self, peak):
        index, matched = binary_search_with_flag(
            self.features, peak.mz, self.error_tolerance)
        return index, matched

    def find_minimizing_index(self, peak, indices):
        best_index = None
        best_error = float('inf')
        for index_case in indices:
            feature = self[index_case]
            err = abs(feature.mz - peak.mz) / peak.mz
            if err < best_error:
                best_index = index_case
                best_error = err
        return best_index

    def handle_peak(self, peak, scan_time):
        if len(self) == 0:
            index = [0]
            matched = False
        else:
            index, matched = self.find_insertion_point(peak)
        if matched:
            feature = self.features[self.find_minimizing_index(peak, index)]
            feature.insert(peak, scan_time)
        else:
            feature = LCMSFeature()
            feature.created_at = "forest"
            feature.insert(peak, scan_time)
            self.insert_feature(feature, index)
        self.count += 1

    def insert_feature(self, feature, index):
        if index[0] != 0:
            self.features.insert(index[0] + 1, feature)
        else:
            if len(self) == 0:
                new_index = index[0]
            else:
                x = self.features[index[0]]
                if x.mz < feature.mz:
                    new_index = index[0] + 1
                else:
                    new_index = index[0]
            self.features.insert(new_index, feature)

    def aggregate_peaks(self, scans, minimum_mz=160, minimum_intensity=500., maximum_mz=float('inf')):
        for scan in scans:
            peak_set = scan.peak_set
            if peak_set is None:
                scan.pick_peaks()
                peak_set = scan.peak_set
            for peak in scan.peak_set:
                if peak.mz < minimum_mz or peak.mz > maximum_mz or peak.intensity < minimum_intensity:
                    continue
                self.handle_peak(peak, scan.scan_time)
        self.features = smooth_overlaps(self.features, self.error_tolerance)

    @classmethod
    def from_reader(cls, reader, error_tolerance=1e-5, minimum_mz=160, minimum_intensity=500., maximum_mz=float('inf'),
                    start_time=None, end_time=None):

        def generate():
            if start_time is not None:
                reader.start_from_scan(rt=start_time, grouped=False)
            else:
                reader.reset()
                reader.make_iterator(grouped=False)
            i = 0
            n = len(reader)
            for i, scan in enumerate(reader):
                if i % 1000 == 0 and i:
                    print("Processed %d/%d Scans (%0.2f%%)" % (i, n, i * 100.0 / n))
                if scan.ms_level == 1:
                    yield scan

        self = cls(error_tolerance=error_tolerance)
        self.aggregate_peaks(
            generate(),
            minimum_mz=minimum_mz,
            minimum_intensity=minimum_intensity,
            maximum_mz=maximum_mz)
        return self


def smooth_overlaps(feature_list, error_tolerance=1e-5):
    feature_list = sorted(feature_list, key=lambda x: x.mz)
    out = []
    last = feature_list[0]
    i = 1
    while i < len(feature_list):
        current = feature_list[i]
        mass_error = abs((last.mz - current.mz) / current.mz)
        if mass_error <= error_tolerance:
            if last.overlaps_in_time(current):
                last = last.merge(current)
                last.created_at = "smooth_overlaps"
            else:
                out.append(last)
                last = current
        else:
            out.append(last)
            last = current
        i += 1
    out.append(last)
    return out


def binary_search_with_flag(array, mz, error_tolerance=1e-5):
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.mz - mz) / mz
        if abs(err) <= error_tolerance:
            i = mid - 1
            # Begin Sweep forward
            while i > 0:
                x = array[i]
                err = (x.mz - mz) / mz
                if abs(err) <= error_tolerance:
                    i -= 1
                    continue
                else:
                    break
            low_end = i
            i = mid + 1

            # Begin Sweep backward
            while i < n:
                x = array[i]
                err = (x.mz - mz) / mz
                if abs(err) <= error_tolerance:
                    i += 1
                    continue
                else:
                    break
            high_end = i
            return list(range(low_end, high_end)), True
        elif (hi - lo) == 1:
            return [mid], False
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0, False


def binary_search(array, mz, error_tolerance=1e-5):
    """Binary search an ordered array of objects with :attr:`mz`
    using a PPM error tolerance of `error_tolerance`

    Parameters
    ----------
    array : list
        An list of objects, sorted over :attr:`mz` in increasing order
    mz : float
        The mz to search for
    error_tolerance : float, optional
        The PPM error tolerance to use when deciding whether a match has been found

    Returns
    -------
    int:
        The index in `array` of the best match
    """
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.mz - mz) / mz
        if abs(err) <= error_tolerance:
            best_index = mid
            best_error = err
            i = mid - 1
            while i >= 0:
                x = array[i]
                err = abs((x.mz - mz) / mz)
                if err < best_error:
                    best_error = err
                    best_index = i
                i -= 1

            i = mid + 1
            while i < n:
                x = array[i]
                err = abs((x.mz - mz) / mz)
                if err < best_error:
                    best_error = err
                    best_index = i
                i += 1
            return best_index
        elif (hi - lo) == 1:
            return None
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0


def search_sweep(array, mz, error_tolerance=1e-5):
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.mz - mz) / mz
        if abs(err) <= error_tolerance:
            i = mid - 1
            while i >= 0:
                x = array[i]
                err = abs((x.mz - mz) / mz)
                if err > error_tolerance:
                    break
                i -= 1
            start = i + 1
            i = mid + 1
            while i < n:
                x = array[i]
                err = abs((x.mz - mz) / mz)
                if err > error_tolerance:
                    break
                i += 1
            end = i
            return (start, end)
        elif (hi - lo) == 1:
            return None
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0


def binary_search_with_flag_neutral(array, neutral_mass, error_tolerance=1e-5):
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - neutral_mass) / neutral_mass
        if abs(err) <= error_tolerance:
            i = mid - 1
            # Begin Sweep forward
            while i > 0:
                x = array[i]
                err = (x.neutral_mass - neutral_mass) / neutral_mass
                if abs(err) <= error_tolerance:
                    i -= 1
                    continue
                else:
                    break
            low_end = i
            i = mid + 1

            # Begin Sweep backward
            while i < n:
                x = array[i]
                err = (x.neutral_mass - neutral_mass) / neutral_mass
                if abs(err) <= error_tolerance:
                    i += 1
                    continue
                else:
                    break
            high_end = i
            return list(range(low_end, high_end)), True
        elif (hi - lo) == 1:
            return [mid], False
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0, False


def binary_search_neutral(array, neutral_mass, error_tolerance=1e-5):
    """Binary search an ordered array of objects with :attr:`neutral_mass`
    using a PPM error tolerance of `error_toler

    Parameters
    ----------
    array : list
        An list of objects, sorted over :attr:`neutral_mass` in increasing order
    neutral_mass : float
        The neutral_mass to search for
    error_tolerance : float, optional
        The PPM error tolerance to use when deciding whether a match has been found

    Returns
    -------
    int:
        The index in `array` of the best match
    """
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - neutral_mass) / neutral_mass
        if abs(err) <= error_tolerance:
            best_index = mid
            best_error = err
            i = mid - 1
            while i >= 0:
                x = array[i]
                err = abs((x.neutral_mass - neutral_mass) / neutral_mass)
                if err < best_error:
                    best_error = err
                    best_index = i
                i -= 1

            i = mid + 1
            while i < n:
                x = array[i]
                err = abs((x.neutral_mass - neutral_mass) / neutral_mass)
                if err < best_error:
                    best_error = err
                    best_index = i
                i += 1
            return best_index
        elif (hi - lo) == 1:
            return None
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0


def search_sweep_neutral(array, neutral_mass, error_tolerance=1e-5):
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - neutral_mass) / neutral_mass
        if abs(err) <= error_tolerance:
            i = mid - 1
            while i >= 0:
                x = array[i]
                err = abs((x.neutral_mass - neutral_mass) / neutral_mass)
                if err > error_tolerance:
                    break
                i -= 1
            start = i + 1
            i = mid + 1
            while i < n:
                x = array[i]
                err = abs((x.neutral_mass - neutral_mass) / neutral_mass)
                if err > error_tolerance:
                    break
                i += 1
            end = i
            return (start, end)
        elif (hi - lo) == 1:
            return None
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0, 0


class DeconvolutedLCMSFeatureMap(object):

    def __init__(self, features):
        self.features = sorted(features, key=lambda x: x.neutral_mass)
        self._by_mz = sorted(features, key=lambda x: x.mz)

    def __len__(self):
        return len(self.features)

    def __iter__(self):
        return iter(self.features)

    def __getitem__(self, i):
        if isinstance(i, (int, slice)):
            return self.features[i]
        else:
            return [self.features[j] for j in i]

    def search(self, mz, error_tolerance=2e-5, use_mz=False):
        if use_mz:
            i = binary_search(self._by_mz, mz, error_tolerance)
            if i is None:
                return None
            return self._by_mz[i]
        i = binary_search_neutral(self.features, mz, error_tolerance)
        if i is None:
            return None
        match = self[i]
        return match

    def find_all(self, mz, error_tolerance=2e-5, use_mz=False):
        if use_mz:
            bounds = search_sweep(self._by_mz, mz, error_tolerance)
            collection = self._by_mz
        else:
            bounds = search_sweep_neutral(self.features, mz, error_tolerance)
            collection = self
        if bounds is not None:
            lo, hi = bounds
            return collection[lo:hi]
        else:
            return []

    def index_range(self, lo, hi, error_tolerance=2e-5):
        return (
            binary_search_with_flag_neutral(
                self.features, lo, error_tolerance)[0][0],
            binary_search_with_flag_neutral(
                self.features, hi, error_tolerance)[0][0])

    def between(self, lo, hi, error_tolerance=2e-5, use_mz=False):
        if use_mz:
            lo_ix = binary_search_with_flag(
                self._by_mz, lo, error_tolerance)[0][0]
            hi_ix = binary_search_with_flag(
                self._by_mz, hi, error_tolerance)[0][0]
            if self._by_mz[lo_ix].mz < lo:
                lo_ix += 1
            if self._by_mz[hi_ix].mz > hi:
                hi_ix -= 1
            if hi_ix < 0:
                hi_ix = 0
            if lo_ix > len(self):
                lo_ix = len(self) - 1
            return self._by_mz[lo_ix:hi_ix]
        else:
            lo_ix = binary_search_with_flag_neutral(
                self.features, lo, error_tolerance)[0][0]
            hi_ix = binary_search_with_flag_neutral(
                self.features, hi, error_tolerance)[0][0]
            if self[lo_ix].neutral_mass < lo:
                lo_ix += 1
            if self[hi_ix].neutral_mass > hi:
                hi_ix -= 1
            if hi_ix < 0:
                hi_ix = 0
            if lo_ix > len(self):
                lo_ix = len(self) - 1
            return self[lo_ix:hi_ix]

    def __repr__(self):
        return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))


def convert_map_to_scan_peak_list(feature_map, peak_loader, time_precision=4, deconvoluted=True):
    metadata_map = {}
    scan_accumulator = defaultdict(list)
    for scan_id, metadata in peak_loader.extended_index.ms1_ids.items():
        metadata_map[round(metadata["scan_time"], time_precision)] = metadata
    for feature in feature_map:
        for node in feature:
            scan_accumulator[round(node.time, time_precision)].extend(node.members)

    packed = []
    for key, peaks in sorted(scan_accumulator.items(), key=lambda x: x[0]):
        template = peak_loader.get_scan_by_time(key)
        if deconvoluted:
            peak_set = PeakSet([])
            deconvoluted_peak_set = DeconvolutedPeakSet(peaks)
        else:
            peak_set = PeakSet(peaks)
            deconvoluted_peak_set = DeconvolutedPeakSet([])
        peak_set.reindex()
        deconvoluted_peak_set.reindex()
        scan = ProcessedScan(
            template.id, template.title, None, template.ms_level, template.scan_time,
            template.index, peak_set, deconvoluted_peak_set, template.polarity,
            None)
        packed.append(scan)
    return packed


class _FeatureIndex(object):
    def __iter__(self):
        return iter(self.features)

    def __getitem__(self, i):
        return self.features[i]

    def __len__(self):
        return len(self.features)

    def __nonzero__(self):
        return bool(self.features)

    def __bool__(self):
        return bool(self.features)

    def __repr__(self):
        return "{self.__class__.__name__}(<{size} features>)".format(self=self, size=len(self))


class MZIndex(_FeatureIndex):
    def __init__(self, features):
        self.features = sorted(features, key=lambda x: x.mz)

    def find_all(self, mz, error_tolerance=2e-5):
        bounds = search_sweep(self.features, mz, error_tolerance)
        if bounds is not None:
            lo, hi = bounds
            return self[lo:hi]
        else:
            return []


class NeutralMassIndex(_FeatureIndex):
    def __init__(self, features):
        self.features = sorted(features, key=lambda x: x.neutral_mass)

    def find_all(self, mass, error_tolerance=2e-5):
        bounds = search_sweep_neutral(self.features, mass, error_tolerance)
        if bounds is not None:
            lo, hi = bounds
            return self[lo:hi]
        else:
            return []


try:
    from ms_deisotope._c.feature_map.feature_map import binary_search_with_flag
except ImportError:
    pass



import io
import gzip

from six import string_types as basestring


GzipFile = _GzipFile = gzip.GzipFile
try:
    import idzip
    GzipFile = idzip.IdzipFile
    WRITE_BUFFER_SIZE = idzip.MAX_MEMBER_SIZE
    has_idzip = True
except (ImportError, AttributeError):
    GzipFile = gzip.GzipFile
    WRITE_BUFFER_SIZE = 2 ** 16
    has_idzip = False


# Do not register idzip with psims, indexing seems to corrupt file
# try:
#     from psims import compression as psims_compression

#     if has_idzip:
#         psims_compression.register(GzipFile, 'gz', b'\037\213')
#         psims_compression.register(GzipFile, 'dz', b'\037\213')
# except ImportError:
#     pass


DEFAULT_BUFFER_SIZE = int(2e6)


def test_gzipped(f):
    """Checks the first two bytes of the
    passed file for gzip magic numbers

    Parameters
    ----------
    f : file-like or path-like

    Returns
    -------
    bool
    """
    if isinstance(f, basestring):
        f = io.open(f, 'rb')
    current = f.tell()
    f.seek(0)
    magic = f.read(2)
    f.seek(current)
    return magic == b'\037\213'


def starts_with_gz_magic(bytestring):
    return bytestring.startswith(b'\037\213')


def get_opener(f, buffer_size=None):
    if buffer_size is None:
        buffer_size = DEFAULT_BUFFER_SIZE
    if not hasattr(f, 'read'):
        f = io.open(f, 'rb')
    buffered_reader = io.BufferedReader(f, buffer_size)
    if test_gzipped(f):
        handle = GzipFile(fileobj=buffered_reader, mode='rb')
    else:
        handle = buffered_reader
    return handle



'''
Deconvolution Pipeline
----------------------

The deconvolution process can be pipelined from start to finish
using the :class:`~.ScanProcessor` class. This includes precursor recalculation and
coisolation detection. :class:`~.ScanProcessor` is intended to be used either as a
replacement for :class:`~.ScanIterator` when deconvolution is desired, or that it's
:meth:`ScanProcessor.process` method will be used to handle individual :class:`~.ScanBunch`
objects/pairs of precursor :class:`~.Scan` objects and a :class:`list` of product :class:`~.Scan`
objects.

The free-function :func:`~.process` is a convenience wrapper around :class:`~.ScanProcessor`,
with fewer configurable parameters.

.. code:: python

    from ms_deisotope import ScanProcessor, glycopeptide, peptide
    from ms_deisotope.scoring import PenalizedMSDeconVFitter, MSDeconVFitter
    from ms_deisotope.test.common import datafile

    # Locate example dataset
    path = datafile("20150710_3um_AGP_001_29_30.mzML.gz")

    proc = processor.ScanProcessor(path, ms1_deconvolution_args={
        "averagine": glycopeptide,
        "scorer": PenalizedMSDeconVFitter(20., 2.),
        "truncate_after": 0.95
    }, msn_deconvolution_args={
        "averagine": peptide,
        "scorer": MSDeconVFitter(10.),
        "truncate_after": 0.8
    })

    bunch = next(proc)
    print(bunch)
    print(bunch.precursor.deconvoluted_peak_set)
'''
import logging

from six import string_types as basestring

import numpy as np

from ms_peak_picker import pick_peaks, PeakSet, PeakIndex
from ms_peak_picker.scan_filter import FTICRBaselineRemoval

from .averagine import AveragineCache, peptide
from .scoring import PenalizedMSDeconVFitter, MSDeconVFitter
from .deconvolution import deconvolute_peaks
from .data_source import MSFileLoader, ScanIterator
from .data_source.common import Scan, ScanBunch, ChargeNotProvided
from .utils import Base
from .peak_dependency_network import NoIsotopicClustersError
from .qc.isolation import PrecursorPurityEstimator
from .task import LogUtilsMixin

logger = logging.getLogger("deconvolution_scan_processor")
logger.addHandler(logging.NullHandler())


def _get_nearest_index(query_mz, peak_list):
    best_index = None
    best_error = float('inf')

    for i, peak in enumerate(peak_list):
        error = abs(peak.mz - query_mz)
        if error < best_error:
            best_error = error
            best_index = i
    return best_index


class PriorityTarget(Base):
    """Represent a targeted envelope deconvolution's parameters and constraints.

    This class is used to tell :func:`ms_deisotope.deconvolution.deconvolute_peaks`
    that the solution produced by this peak should be preferentially extracted.

    Attributes
    ----------
    info : :class:`~.PrecursorInformation`
        The associated precursor information block which contains
        the charge state hint.
    peak : :class:`~.FittedPeak`
        The peak from which to start the deconvolution
    trust_charge_hint : bool
        Whether or not to force the deconvoluter to only consider
        the charge specified in the hint.
    mz : float
        The m/z of :attr:`peak`
    charge : int
        The charge state hint from :attr:`info`
    isolation_window : :class:`~.IsolationWindow`
        The isolation window for this precursor ion. May be `None`
    """

    def __init__(self, peak, info, trust_charge_hint=True, precursor_scan_id=None,
                 product_scan_id=None, isolation_window=None):
        self.peak = peak
        self.info = info
        self.trust_charge_hint = trust_charge_hint
        self.precursor_scan_id = precursor_scan_id
        self.product_scan_id = product_scan_id
        self.isolation_window = isolation_window

    def __iter__(self):
        yield self.peak
        yield self.info

    @property
    def mz(self):
        '''
        The m/z of the matched peak

        Returns
        -------
        float
        '''
        return self.peak.mz

    @property
    def charge(self):
        '''
        The charge state of the precursor ion reported by the source.

        Returns
        -------
        int
        '''
        try:
            return int(self.info.charge)
        except TypeError:
            return 0

    def charge_range_hint(self, charge_range):
        """Create an updated charge range for a Deconvoluter to search.

        At the moment, this only amounts to either returning the charge
        range unchanged or returning a charge range that only contains
        the hinted charge state, depending upon whether :attr:`trust_charge_hint`
        is `False` or not.

        Parameters
        ----------
        charge_range : tuple
            The charge range to update

        Returns
        -------
        tuple
            The updated charge range
        """
        if self.trust_charge_hint and self.info.charge is not ChargeNotProvided:
            return (self.charge, self.charge)
        else:
            return charge_range

    def __repr__(self):
        return "PriorityTarget(mz=%0.4f, intensity=%0.4f, charge_hint=%d)" % (
            self.mz, self.peak.intensity, self.charge)


def _loader_creator(specification):
    if isinstance(specification, basestring):
        return MSFileLoader(specification)
    elif isinstance(specification, ScanIterator):
        return specification
    else:
        raise ValueError("Cannot determine how to get a ScanIterator from %r" % (specification,))


def _simplify_peak_set(peaks, bin_width=5.0):
    bin_edges = np.arange(0, peaks[-1].mz + bin_width, bin_width)
    bins = []
    for i, bin_edge in enumerate(bin_edges, 1):
        if i == len(bin_edges):
            next_edge = bin_edges[-1] + bin_width
        else:
            next_edge = bin_edges[i]
        subset = peaks.between(bin_edge, next_edge)
        bins.append(subset)

    thresholds = []
    reduced_subsets = {}
    k = 0
    for b in bins:
        if len(b) > 0:
            bin_intensities = np.array([p.intensity for p in b])
            thresholds.append(np.max(bin_intensities) / 3.)
            for p in b:
                if p.intensity > thresholds[-1]:
                    reduced_subsets[p.peak_count] = p
            k += (bin_intensities > thresholds[-1]).sum()
        else:
            thresholds.append(0.0)
    subset_peaks = PeakSet(
        sorted(reduced_subsets.values(), key=lambda x: x.mz)).clone()
    subset_peaks.reindex()
    return PeakIndex(np.array([]), np.array([]), subset_peaks)


class ScanProcessor(Base, LogUtilsMixin):
    """Orchestrates the deconvolution of a :class:`~.ScanIterator` scan by scan. This process will
    apply different rules for MS1 scans and MSn scans. This type itself mimics a :class:`~.ScanIterator`,
    consuming (raw) mass spectral data and producing deisotoped and charge deconvolved spectra.

    The algorithms used for each task are independent and can be specified in the appropriate
    attribute dictionary, however there is information sharing between each MS1 scan and its
    MSn scans as the precursor monoisotopic mass is recalibrated according to the MS1 processing
    arguments, and the selected charge state is used to limit the charge range used in the matching
    MSn scan. These are described by :class:`PriorityTarget` objects.

    If an averagine-based deconvoluter is used, the averagine cache will be pre-populated.

    At the moment, MSn assumes only MS2. Until MS3 data become available for testing, this limit
    will remain.

    Attributes
    ----------
    data_source : :class:`str`, :class:`~.ScanIterator` or file-like
        Any valid object to be passed to the `loader_type` callable to produce
        a :class:`~.ScanIterator` instance. A path to a mass spectrometry data file,
        a file-like object, or an instance of :class:`~.ScanIterator`. Used to populate :attr:`reader`
    loader_type : callable
        A callable, which when passed :attr:`data_source` returns an instance of :class:`~.ScanIterator`.
        By default, this is :func:`~.MSFileLoader`. Used to populate :attr:`reader`
    reader: ScanIterator
        Any object implementing the :class:`~.ScanIterator` interface, produced by calling
        :attr:`loader_type` on :attr:`data_source`.
    ms1_deconvolution_args : :class:`dict`
        The arguments passed to :func:`~ms_deisotope.deconvolution.deconvolute_peaks` for MS1
        scans.
    ms1_peak_picking_args : :class:`dict`
        The arguments passed to :func:`ms_peak_picker.pick_peaks` for MS1 scans.
    msn_deconvolution_args : :class:`dict`
        The arguments passed to :func:`~ms_deisotope.deconvolution.deconvolute_peaks` for MSn
        scans.
    msn_peak_picking_args : :class:`dict`
        The arguments passed to :func:`ms_peak_picker.pick_peaks` for MSn scans.
    pick_only_tandem_envelopes : :class:`bool`
        Whether or not to process whole MS1 scans or just the regions around those peaks
        chosen for MSn
    default_precursor_ion_selection_window : :class:`float`
        Size of the selection window to use when :attr:`pick_only_tandem_envelopes` is `True`
        and the information is not available in the scan.
    trust_charge_hint : :class:`bool`
        Whether or not to trust the charge provided by the data source when determining
        the charge state of precursor isotopic patterns. Defaults to `True`
    respect_isolation_window: :class:`bool`
        Whether to use the bounds of the isolation window to reject a monoisotopic peak
        solution
    terminate_on_error: :class:`bool`
        Whether or not  to stop processing on an error. Defaults to `True`
    ms1_averaging: :class:`int`
        The number of adjacent MS1 scans to average prior to picking peaks.
    """

    def __init__(self, data_source, ms1_peak_picking_args=None,
                 msn_peak_picking_args=None,
                 ms1_deconvolution_args=None,
                 msn_deconvolution_args=None,
                 pick_only_tandem_envelopes=False,
                 default_precursor_ion_selection_window=1.5,
                 trust_charge_hint=True,
                 loader_type=None,
                 envelope_selector=None,
                 terminate_on_error=True,
                 ms1_averaging=0,
                 respect_isolation_window=False,
                 too_many_peaks_threshold=7000):
        if loader_type is None:
            loader_type = _loader_creator

        self.data_source = data_source
        self.ms1_peak_picking_args = ms1_peak_picking_args or {}
        self.msn_peak_picking_args = msn_peak_picking_args or ms1_peak_picking_args or {}
        self.ms1_deconvolution_args = ms1_deconvolution_args or {}
        self.ms1_deconvolution_args.setdefault("charge_range", (1, 8))
        self.msn_deconvolution_args = msn_deconvolution_args or {}
        self.msn_deconvolution_args.setdefault("charge_range", (1, 8))
        self.pick_only_tandem_envelopes = pick_only_tandem_envelopes

        self.too_many_peaks_threshold = too_many_peaks_threshold

        self.default_precursor_ion_selection_window = default_precursor_ion_selection_window
        self.respect_isolation_window = respect_isolation_window

        self.trust_charge_hint = trust_charge_hint
        self.ms1_averaging = int(ms1_averaging) if ms1_averaging else 0

        self.loader_type = loader_type

        self._signal_source = self.loader_type(data_source)
        self.envelope_selector = envelope_selector
        self.terminate_on_error = terminate_on_error
        self._prepopulate_averagine_cache()

    def _prepopulate_averagine_cache(self):
        if 'averagine' in self.ms1_deconvolution_args:
            averagine = self.ms1_deconvolution_args['averagine']
            if isinstance(averagine, list):
                averagine = [AveragineCache(a).populate() for a in averagine]
            else:
                averagine = AveragineCache(averagine).populate()
            self.ms1_deconvolution_args['averagine'] = averagine
        if 'averagine' in self.msn_deconvolution_args:
            averagine = self.msn_deconvolution_args['averagine']
            if isinstance(averagine, list):
                averagine = [AveragineCache(a).populate() for a in averagine]
            else:
                averagine = AveragineCache(averagine).populate()
            self.msn_deconvolution_args['averagine'] = averagine

    def _reject_candidate_precursor_peak(self, peak, product_scan):
        isolation = product_scan.isolation_window
        if isolation is None or isolation.is_empty():
            pinfo = product_scan.precursor_information
            err = peak.mz - pinfo.mz
            return abs(err) > self.default_precursor_ion_selection_window
        else:
            return peak.mz not in isolation and self.respect_isolation_window

    @property
    def reader(self):
        '''The :class:`~.ScanIterator` which generates the raw scans that will
        be processed.

        Returns
        -------
        :class:`~.ScanIterator`
        '''
        return self._signal_source

    def _get_envelopes(self, precursor_scan):
        """Get the m/z intervals to pick peaks from for the
        given MS1 scan

        Parameters
        ----------
        precursor_scan: Scan

        Returns
        -------
        list or None
        """
        if not self.pick_only_tandem_envelopes and self.envelope_selector is None:
            return None
        elif self.envelope_selector is None:
            chosen_envelopes = [s.precursor_information for s in precursor_scan.product_scans]
            chosen_envelopes = sorted([(p.mz - 5, p.mz + 10) for p in chosen_envelopes])
        else:
            chosen_envelopes = self.envelope_selector(precursor_scan)
        return chosen_envelopes

    def _pick_precursor_scan_peaks(self, precursor_scan, chosen_envelopes=None):
        """Pick peaks from the given precursor scan

        Parameters
        ----------
        precursor_scan: Scan
            Scan to pick peaks from
        chosen_envelopes: list, optional
            list of m/z intervals to pick peaks for

        Returns
        -------
        PeakSet
        """
        if precursor_scan.is_profile:
            peak_mode = 'profile'
        else:
            peak_mode = 'centroid'
        prec_mz, prec_intensity = precursor_scan.arrays
        if not self.pick_only_tandem_envelopes and self.envelope_selector is None:
            prec_peaks = pick_peaks(prec_mz, prec_intensity, peak_mode=peak_mode, **self.ms1_peak_picking_args)
        else:
            if chosen_envelopes is None:
                chosen_envelopes = self._get_envelopes(precursor_scan)
            prec_peaks = pick_peaks(prec_mz, prec_intensity, peak_mode=peak_mode,
                                    target_envelopes=chosen_envelopes,
                                    **self.ms1_peak_picking_args)
        return prec_peaks

    def _average_ms1(self, precursor_scan):
        """Average signal from :attr:`self.ms1_averaging` scans from
        before and after ``precursor_scan`` and pick peaks from the
        averaged arrays.

        Parameters
        ----------
        precursor_scan: Scan
            The scan to use as a point of reference

        Returns
        -------
        PeakSet
        """
        # averaged scans are always profile mode
        new_scan = precursor_scan.average(self.ms1_averaging)
        prec_peaks = pick_peaks(*new_scan.arrays,
                                target_envelopes=self._get_envelopes(precursor_scan),
                                **self.ms1_peak_picking_args)
        return prec_peaks

    def pick_precursor_scan_peaks(self, precursor_scan):
        """Picks peaks for the given ``precursor_scan`` using the
        appropriate strategy.

        If :attr:`ms1_averaging` > 0, then the signal averaging strategy
        is used, otherwise peaks are picked directly.

        Parameters
        ----------
        precursor_scan: Scan

        Returns
        -------
        PeakSet
        """
        self.log("Picking Precursor Scan Peaks: %r" % (precursor_scan, ))
        if self.ms1_averaging > 0:
            prec_peaks = self._average_ms1(precursor_scan)
        else:
            prec_peaks = self._pick_precursor_scan_peaks(precursor_scan)
        n_peaks = len(prec_peaks)
        if n_peaks > self.too_many_peaks_threshold:
            self.log("%d peaks found for %r, applying local intensity threshold." % (n_peaks, precursor_scan))
            prec_peaks = _simplify_peak_set(prec_peaks)
        precursor_scan.peak_set = prec_peaks
        return prec_peaks

    def pick_product_scan_peaks(self, product_scan):
        """Pick the peaks of product scan

        Parameters
        ----------
        product_scan: :class:`~.Scan`
            The scan to pick peaks from.

        Returns
        -------
        PeakSet
        """
        if product_scan.is_profile:
            peak_mode = 'profile'
        else:
            peak_mode = 'centroid'
        product_mz, product_intensity = product_scan.arrays
        peaks = pick_peaks(product_mz, product_intensity, peak_mode=peak_mode, **self.msn_peak_picking_args)

        if peaks is None:
            raise EmptyScanError(
                "Could not pick peaks for empty product scan", self)

        product_scan.peak_set = peaks
        return peaks

    def get_precursor_peak_for_product_scans(self, precursor_scan):  # pragma: no cover
        """A utility method to obtain :class:`PriorityTarget` objects for
        each product scan of `precursor_scan`.

        Parameters
        ----------
        precursor_scan: :class:`~.Scan`
            The scan to extract.

        Returns
        -------
        :class:`list` of :class:`PriorityTarget`
        """
        priorities = []
        peaks = precursor_scan.peak_set
        for scan in precursor_scan.product_scans:
            precursor_ion = scan.precursor_information
            if peaks is None:
                peaks = self.pick_precursor_scan_peaks(precursor_scan)
            peak, _ = peaks.get_nearest_peak(precursor_ion.mz)
            precursor_ion.peak = peak
            target = PriorityTarget(
                peak,
                precursor_ion,
                self.trust_charge_hint,
                isolation_window=scan.isolation_window)
            if self._reject_candidate_precursor_peak(peak, scan):
                self.log(
                    "Unable to locate a peak for precursor ion %r for tandem scan %s of precursor scan %s" % (
                        precursor_ion, scan.title,
                        precursor_scan.title))
            else:
                priorities.append(target)
        return priorities

    def process_scan_group(self, precursor_scan, product_scans):
        """Performs the initial extraction of information relating
        `precursor_scan` to `product_scans` and picks peaks for ``precursor_scan``.
        Called by :meth:`process`. May be used separately if doing the process step
        by step.

        Parameters
        ----------
        precursor_scan : :class:`~.Scan`
            An MS1 Scan
        product_scans : :class:`list` of :class:`~.Scan`
            A :class:`list` of MSn Scans related to `precursor_scan`

        Returns
        -------
        precursor_scan: :class:`~.Scan`
            As Parameter
        prioritiies: :class:`list` of :class:`~PriorityTarget`
            :class:`list` of the peak target windows in `precursor_scan` which
            are related to `product_scans`
        product_scans: :class:`list` of :class:`~.Scan`
            As Parameter
        """
        prec_peaks = self.pick_precursor_scan_peaks(precursor_scan)
        priorities = []

        if prec_peaks is None:
            raise EmptyScanError(
                "Could not pick peaks for empty precursor scan", self)

        for scan in product_scans:
            precursor_ion = scan.precursor_information
            peak = prec_peaks.has_peak(precursor_ion.mz)
            if peak is not None:
                err = abs(peak.mz - precursor_ion.mz)
            else:
                peak, err = prec_peaks.get_nearest_peak(precursor_ion.mz)
            self.debug("For Precursor at %0.4f, found Peak at %0.4f with error %0.4f" % (
                precursor_ion.mz, peak.mz, err))
            precursor_ion.peak = peak
            target = PriorityTarget(
                peak, precursor_ion, self.trust_charge_hint,
                scan.precursor_information.precursor_scan_id,
                scan.precursor_information.product_scan_id,
                isolation_window=scan.isolation_window)
            if self._reject_candidate_precursor_peak(peak, scan):
                self.log(
                    "Unable to locate a peak for precursor ion %r for tandem scan %s of precursor scan %s" % (
                        precursor_ion, scan.id,
                        precursor_scan.id))
            else:
                priorities.append(target)

        return precursor_scan, priorities, product_scans

    def _default_all_precursor_information(self, scans):
        for scan in scans:
            if scan.ms_level > 1:
                scan.precursor_information.default(orphan=True)

    def deconvolute_precursor_scan(self, precursor_scan, priorities=None):
        """Deconvolute the given precursor scan, giving priority to its product ions,
        correcting the :attr:`precursor_information` attributes of priority targets,
        as well as calculating the degree of precursor purity and coisolating ions.

        Parameters
        ----------
        precursor_scan : :class:`~.Scan`
            The precursor scan to deconvolute
        priorities : :class:`list` of :class:`PriorityTarget`, optional
            The priority targets for the product ions derived from `precursor_scan`

        Returns
        -------
        :class:`~DeconvolutedPeakSet`
            The deconvoluted peaks of ``precursor_scan``
        :class:`list` of :class:`PriorityTarget`
            The precursor ions selected, with updated mass and charge information

        Raises
        ------
        Exception
            Any errors which are thrown during the deconvolution process may be thrown
            if :attr:`terminate_on_error` is `True`.
        """
        if priorities is None:
            priorities = []

        self.log("Deconvoluting Precursor Scan %r" % precursor_scan)
        self.log("Priorities: %r" % priorities)

        ms1_deconvolution_args = self.ms1_deconvolution_args.copy()

        if precursor_scan.polarity in (1, -1):
            polarity = precursor_scan.polarity
            ms1_deconvolution_args['charge_range'] = tuple(
                polarity * abs(c) for c in ms1_deconvolution_args['charge_range'])
        try:
            decon_result = deconvolute_peaks(
                precursor_scan.peak_set, priority_list=priorities,
                **ms1_deconvolution_args)
        except NoIsotopicClustersError as e:
            e.scan_id = precursor_scan.id
            if self.terminate_on_error:
                raise e
            else:
                self.log("No isotopic clusters found in %r" % precursor_scan.id)

        dec_peaks, priority_results = decon_result
        if decon_result.errors:
            self.error("Errors occurred during deconvolution of %s, %r" % (
                precursor_scan.id, decon_result.errors))
        precursor_scan.deconvoluted_peak_set = dec_peaks
        for pr in priority_results:
            if pr is None:
                continue
            else:
                pr.chosen_for_msms = True

        # `priorities` and `priority_results` are parallel lists. The
        # ith position in `priorities` corresponds to the ith deconvoluted
        # priority result in `priority_results`. The entry in `priority_results`
        # may be `None` if the deconvolution failed, but elements of `priorities`
        # should always be FittedPeak or Peak-like instances

        coisolation_detection = PrecursorPurityEstimator(default_width=self.default_precursor_ion_selection_window)
        self.debug("Priority Targets for %s: %r" % (
            precursor_scan.id, [
                (p.mz, p.charge) if p is not None else None for p in priorities
            ]))
        for product_scan in precursor_scan.product_scans:
            precursor_information = product_scan.precursor_information

            # unknown precursor purity
            product_scan.annotations['precursor purity'] = 0.0
            i = _get_nearest_index(precursor_information.mz, priorities)

            # If no peak is found in the priority list, it means the priority list is empty.
            # This should never happen in the current implementation. If it did, then we forgot
            # to pass the priority list to this function.
            if i is None:
                self.log(
                    "Could not find deconvolution for %r (No nearby peak in the priority list)" %
                    precursor_information)
                precursor_information.default(orphan=True)
                continue

            peak = priority_results[i]
            # If the deconvolution result is None, then we have no answer
            if peak is None:
                self.log(
                    "Could not find deconvolution for %r (No solution was found for this region)" %
                    precursor_information)
                precursor_information.default(orphan=True)

                continue
            elif peak.charge == 1 or (peak.charge != precursor_information.charge and self.trust_charge_hint):
                if precursor_information.charge != ChargeNotProvided:
                    self.log(
                        "Could not find deconvolution for %r (Unacceptable solution was proposed: %r)" %
                        (precursor_information, peak))
                    precursor_information.default()
                    continue

            precursor_purity = -1.0
            if peak is not None:
                precursor_purity, coisolation = coisolation_detection(
                    precursor_scan,
                    peak,
                    product_scan.isolation_window)
                precursor_information.coisolation = coisolation
                self.debug(
                    "Precursor m/z %f\nExperimental = %r\nTheoretical = %r" % (
                        peak.mz,
                        ', '.join(["(%0.4f, %0.1f)" % (p.mz, p.intensity) for p in peak.envelope]),
                        ', '.join(["(%0.4f, %0.1f)" % (p.mz, p.intensity) for p in peak.fit.theoretical]))
                )

            product_scan.annotations['precursor purity'] = precursor_purity
            precursor_information.extract(peak)
        return dec_peaks, priority_results

    def deconvolute_product_scan(self, product_scan):
        """Deconvolute the peaks of `product_scan`.

        This method will override the upper limit "charge_range" of
        :attr:`msn_deconvolution_args` to the charge information of
        the precursor ion.

        This method sets the :attr:`~.Scan.deconvoluted_peak_set` of
        `product_scan`.

        Parameters
        ----------
        product_scan : :class:`~.Scan`
            The scan to deconvolute.

        Returns
        -------
        :class:`~.DeconvolutedPeakSet`

        Raises
        ------
        Exception
            Any errors which are thrown during the deconvolution process may be thrown
            if :attr:`terminate_on_error` is `True`.
        """
        self.log("Deconvoluting Product Scan %r" % (product_scan, ))
        precursor_ion = product_scan.precursor_information
        top_charge_state = precursor_ion.extracted_charge
        if not top_charge_state:
            top_charge_state = precursor_ion.charge
        deconargs = dict(self.msn_deconvolution_args)
        charge_range = list(deconargs.get("charge_range", [1, top_charge_state]))
        if top_charge_state is not None and top_charge_state is not ChargeNotProvided and\
           top_charge_state != 0 and abs(top_charge_state) < abs(charge_range[1]):
            charge_range[1] = top_charge_state

        deconargs["charge_range"] = charge_range

        if product_scan.polarity in (-1, 1):
            polarity = product_scan.polarity
            deconargs["charge_range"] = [
                polarity * abs(c) for c in deconargs["charge_range"]]

        try:
            dec_peaks, _ = deconvolute_peaks(product_scan.peak_set, **deconargs)
        except NoIsotopicClustersError as e:
            self.log("No Isotopic Clusters found in %r" % product_scan.id)
            e.scan_id = product_scan.id
            if self.terminate_on_error:
                raise e

        product_scan.deconvoluted_peak_set = dec_peaks
        return dec_peaks

    def _get_next_scans(self):
        bunch = next(self.reader)
        try:
            precursor, products = bunch
        except ValueError:
            if isinstance(bunch, Scan):
                if bunch.ms_level == 1:
                    precursor = bunch
                    products = []
                else:
                    precursor = None
                    products = [bunch]

        if self.pick_only_tandem_envelopes:
            while len(products) == 0:
                precursor, products = next(self.reader)

        return precursor, products

    def process(self, precursor, products):
        """Fully preprocesses the `precursor` and `products` scans, performing
        any necessary information sharing.

        This method may be used to process scans from other sources not from the
        wrapped :class:`~.ScanIterator`.

        Parameters
        ----------
        precursor : :class:`~.Scan`
            An MS1 Scan
        products : :class:`list` of :class:`~.Scan`
            A list of MSn Scans related to `precursor`

        Returns
        -------
        :class:`~.ScanBunch`
            The fully processed version of `precursor` and `products`
        """
        precursor_scan, priorities, product_scans = self.process_scan_group(precursor, products)
        if precursor_scan is not None:
            self.deconvolute_precursor_scan(precursor_scan, priorities)
        else:
            self._default_all_precursor_information(product_scans)

        for product_scan in product_scans:
            self.pick_product_scan_peaks(product_scan)
            self.deconvolute_product_scan(product_scan)

        return ScanBunch(precursor_scan, product_scans)

    def next(self):
        """Fetches the next bunch of scans from :attr:`reader` and
        invokes :meth:`process` on them, picking peaks and deconvoluting them.

        Returns
        -------
        ScanBunch
        """
        precursor, products = self._get_next_scans()
        bunch = self.process(precursor, products)
        return bunch

    def __next__(self):
        """Fetches the next bunch of scans from :attr:`reader` and
        invokes :meth:`process` on them, picking peaks and deconvoluting them.

        Returns
        -------
        ScanBunch
        """
        return self.next()

    def __iter__(self):
        return self

    def pack_next(self):
        """As :meth:`next`, except instead of producing :class:`ScanBunch` of
        :class:`Scan` instances, instead it uses :class:`ProcessedScan` to strip away
        much of the heavy information like the raw data arrays.

        Returns
        -------
        ScanBunch
        """
        precursor, products = self._get_next_scans()
        precursor_scan, product_scans = self.process(precursor, products)
        return ScanBunch(precursor_scan.pack() if precursor_scan else None, [p.pack() for p in product_scans])

    def start_from_scan(self, *args, **kwargs):
        """A wrapper around :meth:`~.RandomAccessScanSource.start_from_scan` provided by
        :attr:`reader`, if available.

        Returns
        -------
        self

        See Also
        --------
        :meth:`~.RandomAccessScanSource.start_from_scan`
        """
        self.reader.start_from_scan(*args, **kwargs)
        return self


ScanProcessor.log_with_logger(logger)


def process(data_source, ms1_averagine=peptide, msn_averagine=peptide,
            ms1_score_threshold=20, msn_score_threshold=5, denoise=False,
            ms1_max_missed_peaks=1, pick_only_tandem_envelopes=False,
            trust_charge_hint=True, envelope_selector=None, terminate_on_error=True,
            ms1_averaging=0, respect_isolation_window=False, use_quick_charge=True):
    """Construct a deconvolution pipeline for common applications.

    Parameters
    ----------
    data_source : :class:`str` or :class:`~.ScanIterator` or file-like object
        The scan data source to read raw spectra from
    ms1_averagine : :class:`~.Averagine` or :class:`~.AveragineCache`, optional
        The :class:`~.Averagine` model to use for MS1 scans. Defaults to
        :data:`ms_deisotope.averagine.peptide`.
    msn_averagine : :class:`~.Averagine` or :class:`~.AveragineCache`, optional
        The :class:`~.Averagine` model to use for MSn scans. Defaults to
        :data:`ms_deisotope.averagine.peptide`.
    ms1_score_threshold : float, optional
        The score threshold to use to reject isotopic pattern fits for MS1 scans.
        The default is 20.0.
    msn_score_threshold : float, optional
        The score threshold to use to reject isotopic pattern fits for MS1 scans.
        The default is 5.0.
    denoise : :class:`bool` or :class:`float`, optional
        Whether to denoise MS1 scans. If the value is not false-y, it may either be
        a float to set the scale of the denoising process, or 5.0 if the value is
        :const:`True`.
    ms1_max_missed_peaks : :class:`int`, optional
        The maximum number of missed peaks to permit for MS1 scans. The default is 1.
    pick_only_tandem_envelopes : :class:`bool`
        Whether or not to process whole MS1 scans or just the regions around those peaks
        chosen for MSn
    default_precursor_ion_selection_window : :class:`float`
        Size of the selection window to use when :attr:`pick_only_tandem_envelopes` is `True`
        and the information is not available in the scan.
    trust_charge_hint : :class:`bool`
        Whether or not to trust the charge provided by the data source when determining
        the charge state of precursor isotopic patterns. Defaults to `True`
    terminate_on_error: :class:`bool`
        Whether or not  to stop processing on an error. Defaults to `True`
    ms1_averaging: :class:`int`
        The number of adjacent MS1 scans to average prior to picking peaks.
    respect_isolation_window: :class:`bool`
        Whether to use the bounds of the isolation window to reject a monoisotopic peak
        solution
    use_quick_charge : :class:`bool`, optional
        Whether or not to used the QuickCharge algorithm for expediting charge calculation.

    Returns
    -------
    :class:`ScanProcessor`
    """
    if denoise:
        ms1_peak_picking_args = {
            "filters": [
                FTICRBaselineRemoval(
                    scale=denoise if denoise is not True else 5., window_length=2)
            ]
        }
    else:
        ms1_peak_picking_args = None

    ms1_deconvolution_args = {
        "averagine": ms1_averagine,
        "scorer": PenalizedMSDeconVFitter(ms1_score_threshold, 2.0),
        "use_quick_charge": use_quick_charge,
        "max_missed_peaks": ms1_max_missed_peaks,
        "truncate_after": 0.95,
    }
    msn_deconvolution_args = {
        "averagine": msn_averagine,
        "scorer": MSDeconVFitter(msn_score_threshold),
        "use_quick_charge": use_quick_charge,
        "truncate_after": 0.8,
    }
    processor = ScanProcessor(
        data_source, ms1_peak_picking_args,
        None, ms1_deconvolution_args,
        msn_deconvolution_args, pick_only_tandem_envelopes,
        trust_charge_hint=trust_charge_hint,
        envelope_selector=envelope_selector,
        terminate_on_error=terminate_on_error,
        ms1_averaging=ms1_averaging,
        respect_isolation_window=respect_isolation_window)
    return processor


class EmptyScanError(ValueError):
    """A sub-type of :class:`ValueError` which is used to indicate
    that a spectrum is empty and could not be manipulated.
    """
    def __init__(self, msg, scan_id=None):
        ValueError.__init__(self, msg)
        self.scan_id = scan_id



import unittest

from ms_deisotope.test.common import datafile
from ms_deisotope.data_source import infer_type
from ms_deisotope.data_source.metadata import scan_traits


class TestScanTraits(unittest.TestCase):
    path = datafile("three_test_scans.mzML")

    @property
    def reader(self):
        return infer_type.MSFileLoader(self.path)

    def test_traits(self):
        bunch = next(self.reader)
        scan = bunch.precursor
        acquisition = scan.acquisition_information
        assert len(acquisition) == 1
        scan_event = acquisition[0]
        assert not scan_event.has_ion_mobility()
        assert len(scan_event) == 1
        scan_window = scan_event[0]
        assert scan_window.lower == 350
        assert scan_window.upper == 1500
        assert not scan_window.is_empty()
        assert scan_traits.ScanWindow(0, 0).is_empty()
        assert scan_traits.ScanWindow(350, 1500) == scan_window
        assert scan_window == scan_event.total_scan_window()
        assert scan_event == scan_event
        assert acquisition == acquisition
        assert scan.tic.raw() - 1.8161617e+10 + 104 == 0.0
        scan.pick_peaks()
        assert abs(scan.tic.centroided() - 4531158140.658203) < 1e-3
        assert abs(scan.tic() - 4531158140.658203) < 1e-3

        scan = bunch.products[0]
        isolation = scan.isolation_window
        assert not isolation.is_empty()
        assert scan_traits.IsolationWindow(None, 200, None).is_empty()
        assert scan_traits.IsolationWindow(0, 200, 0).is_empty()
        assert isolation.spans(isolation.target)
        assert isolation == isolation





'''A set of strategies for retaining peaks following deconvolution to
recover from low quality spectra. These methods are not recommended for
use on spectra where most ions are multiply charged.

These objects are meant to be provided to :func:`~.deconvolut_peaks`'s
`retention_strategy` argument, which will take care of calling them with
the appropriate arguments.
'''
import operator
import abc

from six import add_metaclass

from ms_deisotope.averagine import neutral_mass

from .utils import from_fitted_peak

intensity_getter = operator.attrgetter("intensity")


@add_metaclass(abc.ABCMeta)
class PeakRetentionStrategyBase(object):
    """A strategy for retaining peaks left behind by the deconvolution procedure.

    Deconvolution requires that the isotopic peak structure is present in the peak list,
    which may not be consistent for all types of data. This type of method is appropriate
    when the caller expects real peaks may have been left behind, as is the case with smaller
    peptides and lower quality tandem mass spectra.
    """

    def create_peak(self, fitted_peak, charge=1):
        """Create a :class:`~.DeconvolutedPeak` from a given :class:`~.FittedPeak` ``peak``
        and an optional charge state.

        Parameters
        ----------
        fitted_peak : :class:`~.FittedPeak`
            The peak to treat as the monoisotopic peak of the unfitted deconvoluted peak
        charge : int, optional
            The charge state to specify the peak at (the default is 1)

        Returns
        -------
        DeconvolutedPeak
        """
        return from_fitted_peak(fitted_peak, charge)

    @abc.abstractmethod
    def retain_peaks(self, peaklist, original_peaklist=None, charge_range=None, solutions=None):
        """Given a list of :class:`~.FittedPeak` objects left over after deconvolution,
        produce a list of :class:`~.DeconvolutedPeak` objects from them according to some
        criteria.

        Deconvolution requires that the isotopic peak structure is present in the peak list,
        which may not be consistent for all types of data. This type of method is appropriate
        when the caller expects real peaks may have been left behind, as is the case with smaller
        peptides and lower quality tandem mass spectra.

        Parameters
        ----------
        peaklist : :class:`~.PeakSet`
            The list of peaks that remain after deconvolution
        original_peaklist : :class:`~.PeakSet`, optional
            The list of peaks that were initially presented for deconvolution,
            which may be used to infer relative thresholds from. If not provided,
            these thresholds may be learned from the left over peaks, but the
            parameters may not be as good.
        charge_range : :class:`tuple`, optional
            The range of charge states considered during deconvolution. Used to infer
            the minimum charge state to assign to the peaks this method creates. If
            not provided, the default charge state is :const`1`.
        solutions: :class:`~.DeconvolutedPeakSet`, optional
            The accepted solutions which might be used to infer/reject additional peaks.

        Returns
        -------
        :class:`list`:
            The list of :class:`~.DeconvolutedPeak` objects.
        """
        raise NotImplementedError()

    def __call__(self, peaklist, original_peaklist=None, charge_range=None, solutions=None):
        """A wrapper for :meth:`retain_peaks`. Given a list of :class:`~.FittedPeak`
        objects left over after deconvolution, produce a list of :class:`~.DeconvolutedPeak`
        objects from them according to some criteria.

        Parameters
        ----------
        peaklist : :class:`~.PeakSet`
            The list of peaks that remain after deconvolution
        original_peaklist : :class:`~.PeakSet`, optional
            The list of peaks that were initially presented for deconvolution,
            which may be used to infer relative thresholds from. If not provided,
            these thresholds may be learned from the left over peaks, but the
            parameters may not be as good.
        charge_range : :class:`tuple`, optional
            The range of charge states considered during deconvolution. Used to infer
            the minimum charge state to assign to the peaks this method creates. If
            not provided, the default charge state is :const`1`.
        solutions: :class:`~.DeconvolutedPeakSet`, optional
            The accepted solutions which might be used to infer/reject additional peaks.

        Returns
        -------
        :class:`list`:
            The list of :class:`~.DeconvolutedPeak` objects.

        See Also
        --------
        :meth:`retain_peaks`
        """
        return self.retain_peaks(peaklist, original_peaklist, charge_range)

    def infer_minimum_charge(self, charge_range):
        """Given a charge range :class:`tuple`, return the smallest absolute magnitude
        charge state possible in that range.

        This method is polarity aware, and does its reasoning on the absolute magnitude
        of the charge state range, rather than on its signed value.

        Parameters
        ----------
        charge_range : :class:`tuple`
            The range of charge values that were used during the deconvolution process

        Returns
        -------
        :class:`int`:
            The minimum charge state.
        """
        if charge_range is None:
            return 1
        abs_range = [abs(c) for c in charge_range]
        if abs_range[0] < abs_range[1]:
            min_charge = charge_range[0]
        else:
            min_charge = charge_range[1]
        return min_charge


class TopNRetentionStrategy(PeakRetentionStrategyBase):
    """This strategy retains up to at most :attr:`n_peaks` peaks from the
    leftover signal, and requires that any peaks it retains are atleast
    :attr:`base_peak_coefficient` * the base peak of the original peak list.

    This strategy treats the leftover peak as if it were the monoisotopic peak
    of an unobserved isotopic distribution. Because the monoisotopic peak ceases
    to dominate an isotopic distribution above a certain mass (as implied by an
    averagine isotopic model), any peak selected must have a neutral mass below
    :attr:`max_mass`.

    Attributes
    ----------
    n_peaks : int
        The maximum number of peaks to retain.
    base_peak_coefficient : float
        The fraction of the base peak intensity to threshold on. Defaults to :const:`0.05`
    max_mass : float
        The largest neutral mass that a peak may have before it would no longer have a dominant
        monoisotopic peak and the peak will be discarded.

    """
    def __init__(self, n_peaks=50, base_peak_coefficient=0.05, max_mass=850.0):
        self.n_peaks = n_peaks
        self.base_peak_coefficient = base_peak_coefficient
        self.max_mass = max_mass

    def retain_peaks(self, peaklist, original_peaklist=None, charge_range=None, solutions=None):
        if original_peaklist is None:
            base_peak_sequence = peaklist
        else:
            base_peak_sequence = original_peaklist
        try:
            base_peak = max([peak.intensity for peak in base_peak_sequence])
        except ValueError:
            return []
        min_charge = self.infer_minimum_charge(charge_range)
        min_charge /= abs(min_charge)
        threshold = self.base_peak_coefficient * base_peak
        peaklist = sorted(peaklist, key=intensity_getter, reverse=True)
        result = []
        for peak in peaklist:
            if neutral_mass(peak.mz, min_charge) > self.max_mass:
                continue
            if peak.intensity >= threshold:
                result.append(self.create_peak(peak, min_charge))
                if len(result) == self.n_peaks:
                    break
            else:
                break

        return result


simple_peak_retention = TopNRetentionStrategy()
